r"""
Quantum-valued polynomial rings

This provides a `q`-analogue of the :class:`~sage.rings.polynomials.integer_valued_polynomials.IntegerValuedPolynomialRing`.

AUTHORS:

- Frédéric Chapoton (2024-03): Initial version
"""
# ****************************************************************************
#  Copyright (C) 2024 Frédéric Chapoton <chapoton-math-unistra-fr>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from sage.arith.misc import binomial
from sage.categories.algebras import Algebras
from sage.categories.realizations import Category_realization_of_parent
from sage.categories.rings import Rings
from sage.combinat.free_module import CombinatorialFreeModule
from sage.combinat.q_analogues import q_binomial, q_int
from sage.data_structures.blas_dict import linear_combination
from sage.matrix.constructor import matrix
from sage.misc.bindable_class import BindableClass
from sage.misc.cachefunc import cached_method
from sage.modules.free_module_element import vector
from sage.rings.integer_ring import ZZ
from sage.rings.polynomial.laurent_polynomial_ring import LaurentPolynomialRing
from sage.rings.polynomial.polynomial_ring import polygen
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
from sage.rings.rational_field import QQ
from sage.sets.family import Family
from sage.sets.non_negative_integers import NonNegativeIntegers
from sage.structure.element import parent
from sage.structure.parent import Parent
from sage.structure.unique_representation import UniqueRepresentation


def q_int_x(n, q=None):
    """
    Return the interpolating polynomial of `q`-integers.

    INPUT:

    - ``n`` -- a positive integer

    - ``q`` -- optional variable

    EXAMPLES::

        sage: from sage.rings.polynomial.q_integer_valued_polynomials import q_int_x
        sage: q_int_x(3)
        q^2*x + q + 1

    TESTS::

        sage: from sage.rings.polynomial.q_integer_valued_polynomials import q_int_x
        sage: q_int_x(3, 1)
        x + 2
    """
    if q is None:
        ring_q = PolynomialRing(ZZ, 'q')
        q = ring_q.gen()
    else:
        ring_q = q.parent()
    x = polygen(ring_q, 'x')
    return q_int(n - 1, q) + q**(n - 1) * x


def q_binomial_x(m, n, q=None):
    r"""
    Return a `q`-analogue of ``binomial(m + x, n)``.

    When evaluated at the `q`-integer `[k]_q`, this gives
    the usual `q`-binomial coefficient `[m + k, n]_q`.

    INPUT:

    - ``m`` and ``n`` -- positive integers

    - ``q`` -- optional variable

    EXAMPLES::

        sage: from sage.combinat.q_analogues import q_int
        sage: from sage.rings.polynomial.q_integer_valued_polynomials import q_binomial_x, q_int_x
        sage: q_binomial_x(4,2)(0) == q_binomial(4,2)
        True
        sage: q_binomial_x(3,2)(1) == q_binomial(4,2)
        True
        sage: q_binomial_x(3,1) == q_int_x(4)
        True
        sage: q_binomial_x(2,0).parent()
        Univariate Polynomial Ring in x over Fraction Field of
        Univariate Polynomial Ring in q over Integer Ring

    TESTS::

        sage: from sage.rings.polynomial.q_integer_valued_polynomials import q_binomial_x
        sage: q_binomial_x(4,2,1)
        1/2*x^2 + 7/2*x + 6
    """
    if q is None:
        ring_q = PolynomialRing(ZZ, 'q')
    else:
        ring_q = q.parent()
    ring = PolynomialRing(ring_q.fraction_field(), 'x')
    if n == 0:
        return ring.one()
    return ring.prod(q_int_x(m + 2 - i, q) / q_int(i, q)
                     for i in range(1, n + 1))


class QuantumValuedPolynomialRing(UniqueRepresentation, Parent):
    r"""
    The quantum-valued polynomial ring over a base ring.

    Quantum-valued polynomial rings are commutative and associative
    algebras, with a basis indexed by nonnegative integers.

    The elements are polynomials in one variable `x` with coefficients in
    the field of rational functions in `q`, such that the value at
    every nonnegative `q`-integer is a polynomial in `q`.

    This algebra is endowed with two bases, named ``B`` or ``Binomial``
    and ``S`` or ``Shifted``.

    INPUT:

    - ``R`` -- commutative ring

    - ``q`` -- optional variable

    There are two possible input formats:

    - If the argument ``q`` is not given, then the ring ``R`` is
      taken as a base ring and the ring of Laurent polynomials in `q`
      over ``R`` is built and used.

    - If the argument ``q`` is given, then it should belong to the ring ``R``
      and be invertible in this ring.

    EXAMPLES::

        sage: F = QuantumValuedPolynomialRing(QQ).S(); F
        Quantum-Valued Polynomial Ring over Rational Field
        in the shifted basis

        sage: F.gen()
        S[1]

        sage: S = QuantumValuedPolynomialRing(ZZ); S
        Quantum-Valued Polynomial Ring over Integer Ring
        sage: S.base_ring()
        Univariate Laurent Polynomial Ring in q over Integer Ring

    Quantum-valued polynomial rings commute with their base ring::

        sage: K = QuantumValuedPolynomialRing(QQ).S()
        sage: a = K.gen()
        sage: c = K.monomial(2)

    Quantum-valued polynomial rings are commutative::

        sage: c^3 * a == c * a * c * c
        True

    We can also manipulate elements in the basis and coerce elements from our
    base field::

        sage: F = QuantumValuedPolynomialRing(QQ).S()
        sage: B = F.basis()
        sage: B[2] * B[3]
        (q^-5+q^-4+q^-3)*S[3] - (q^-6+2*q^-5+3*q^-4+3*q^-3+2*q^-2+q^-1)*S[4]
        + (q^-6+q^-5+2*q^-4+2*q^-3+2*q^-2+q^-1+1)*S[5]
        sage: 1 - B[2] * B[2] / 2
        S[0] - (1/2*q^-3)*S[2] + (1/2*q^-4+q^-3+q^-2+1/2*q^-1)*S[3]
        - (1/2*q^-4+1/2*q^-3+q^-2+1/2*q^-1+1/2)*S[4]
    """
    @staticmethod
    def __classcall_private__(cls, R, q=None) -> None:
        """
        Normalize the input.

        EXAMPLES::

            sage: q = LaurentPolynomialRing(QQ, 'q').gen()
            sage: F1 = QuantumValuedPolynomialRing(QQ)
            sage: F2 = QuantumValuedPolynomialRing(q.parent(), q)
            sage: F1 is F2
            True
        """
        if R not in Rings().Commutative():
            msg = "argument R must be a commutative ring"
            raise TypeError(msg)

        if q is None:
            laurent = LaurentPolynomialRing(R, 'q')
            q = laurent.gen()
        else:
            laurent = q.parent()
        return super().__classcall__(cls, laurent, q)

    def __init__(self, R, q) -> None:
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: F = QuantumValuedPolynomialRing(QQ); F
            Quantum-Valued Polynomial Ring over Rational Field
            sage: TestSuite(F).run()

            sage: F = QuantumValuedPolynomialRing(QQ, 1); F
            Quantum-Valued Polynomial Ring over Integer Ring
            sage: TestSuite(F).run()

        TESTS::

            sage: QuantumValuedPolynomialRing(24)
            Traceback (most recent call last):
            ...
            TypeError: argument R must be a commutative ring
        """
        self._ground_ring = R.base_ring()
        self._q = q
        cat = Algebras(R).Commutative().WithBasis()
        Parent.__init__(self, base=R, category=cat.WithRealizations())

    _shorthands = ("B", "S")

    def _repr_(self) -> str:
        r"""
        Return the string representation.

        EXAMPLES::

            sage: F = QuantumValuedPolynomialRing(QQ)
            sage: F  # indirect doctest
            Quantum-Valued Polynomial Ring over Rational Field

            sage: QuantumValuedPolynomialRing(ZZ)
            Quantum-Valued Polynomial Ring over Integer Ring
        """
        base = self.base_ring().base_ring()
        return f"Quantum-Valued Polynomial Ring over {base}"

    def a_realization(self):
        """
        Return a default realization.

        The Shifted realization is chosen.

        EXAMPLES::

            sage: QuantumValuedPolynomialRing(QQ).a_realization()
            Quantum-Valued Polynomial Ring over Rational Field
            in the shifted basis
        """
        return self.Shifted()

    class Bases(Category_realization_of_parent):
        def super_categories(self) -> list:
            r"""
            Return the super-categories of ``self``.

            EXAMPLES::

                sage: A = QuantumValuedPolynomialRing(QQ); A
                Quantum-Valued Polynomial Ring over Rational Field
                sage: C = A.Bases(); C
                Category of bases of Quantum-Valued Polynomial Ring
                over Rational Field
                sage: C.super_categories()
                [Category of realizations of Quantum-Valued Polynomial Ring
                 over Rational Field,
                 Join of Category of algebras with basis
                 over Univariate Laurent Polynomial Ring in q over Rational Field and
                 Category of filtered algebras
                 over Univariate Laurent Polynomial Ring in q over Rational Field and
                 Category of commutative algebras
                 over Univariate Laurent Polynomial Ring in q over Rational Field and
                 Category of realizations of unital magmas]
            """
            A = self.base()
            category = Algebras(A.base_ring()).Commutative().Filtered()
            return [A.Realizations(),
                    category.Realizations().WithBasis()]

        class ParentMethods:
            def ground_ring(self):
                """
                Return the ring of coefficients.

                This ring is not supposed to contain the variable `q`.

                EXAMPLES::

                    sage: A = QuantumValuedPolynomialRing(QQ).S()
                    sage: A.ground_ring()
                    Rational Field
                """
                return self.realization_of()._ground_ring

            def _repr_(self) -> str:
                r"""
                EXAMPLES::

                    sage: F = QuantumValuedPolynomialRing(QQ).S()
                    sage: F  # indirect doctest
                    Quantum-Valued Polynomial Ring over Rational Field
                    in the shifted basis
                """
                real = self.realization_of()
                return f"{real} in the {self._realization_name()} basis"

            @cached_method
            def one_basis(self):
                r"""
                Return the number 0, which index the unit of this algebra.

                EXAMPLES::

                    sage: A = QuantumValuedPolynomialRing(QQ).S()
                    sage: A.one_basis()
                    0
                    sage: A.one()
                    S[0]
                """
                return self.basis().keys()(0)

            def q(self):
                """
                Return the variable `q`.

                EXAMPLES::

                    sage: A = QuantumValuedPolynomialRing(QQ).S()
                    sage: A.q()
                    q
                """
                return self.realization_of()._q

            def degree_on_basis(self, m):
                r"""
                Return the degree of the basis element indexed by ``m``.

                EXAMPLES::

                    sage: A = QuantumValuedPolynomialRing(QQ).S()
                    sage: A.degree_on_basis(4)
                    4
                """
                return ZZ(m)

            def from_polynomial(self, p):
                r"""
                Convert a polynomial into the ring of quantum-valued polynomials.

                This raises a :exc:`ValueError` if this is not possible.

                INPUT:

                - ``p`` -- a polynomial in ``x`` with coefficients in ``QQ(q)``

                EXAMPLES::

                    sage: A = QuantumValuedPolynomialRing(ZZ).S()
                    sage: S = A.basis()
                    sage: A.from_polynomial((S[1]).polynomial())
                    S[1]
                    sage: A.from_polynomial((S[2]+2*S[3]).polynomial())
                    S[2] + 2*S[3]

                    sage: A = QuantumValuedPolynomialRing(ZZ).B()
                    sage: B = A.basis()
                    sage: A.from_polynomial((B[1]).polynomial())
                    B[1]
                    sage: A.from_polynomial((B[2]+2*B[3]).polynomial())
                    B[2] + 2*B[3]

                TESTS::

                    sage: A = QuantumValuedPolynomialRing(QQ).B()
                    sage: q = polygen(QQ,'q')
                    sage: x = polygen(q.parent(),'x')
                    sage: A.from_polynomial(x**2/(q+1)+1)
                    Traceback (most recent call last):
                    ...
                    ValueError: not a polynomial with integer values : 1/(q + 1) is not a Laurent polynomial
                """
                B = self.basis()
                poly = self._poly
                laurent_polys = self.base_ring()
                remain = p.change_variable_name('x')
                result = self.zero()
                while remain:
                    N = remain.degree()
                    base_N = poly(N)
                    top_coeff = remain.leading_coefficient() / base_N.lc()
                    denom = top_coeff.denominator()
                    if denom.is_term():
                        numer = top_coeff.numerator()
                        top_coeff_laurent = laurent_polys(numer) / laurent_polys(denom)
                    else:
                        msg = 'not a polynomial with integer values :'
                        msg += f' {top_coeff} is not a Laurent polynomial'
                        raise ValueError(msg)
                    remain += -top_coeff * base_N
                    result += top_coeff_laurent * B[N]
                return result

            def gen(self, i=0):
                r"""
                Return the generator of the algebra.

                The optional argument is ignored.

                EXAMPLES::

                    sage: F = QuantumValuedPolynomialRing(ZZ).S()
                    sage: F.gen()
                    S[1]
                """
                return self.basis()[1]

            @cached_method
            def algebra_generators(self):
                r"""
                Return the generators of this algebra.

                EXAMPLES::

                    sage: A = QuantumValuedPolynomialRing(ZZ).S(); A
                    Quantum-Valued Polynomial Ring over Integer Ring
                    in the shifted basis
                    sage: A.algebra_generators()
                    Family (S[1],)
                """
                return Family([self.basis()[1]])

            gens = algebra_generators

        class ElementMethods:
            def __call__(self, v):
                """
                Return the evaluation at some value ``v``.

                EXAMPLES::

                     sage: F = QuantumValuedPolynomialRing(ZZ).S()
                     sage: B = F.gen()
                     sage: f = B**2+4*B+6
                     sage: f(1/3)
                     (q^2 + 18*q + 99)/9

                     sage: F = QuantumValuedPolynomialRing(ZZ).B()
                     sage: B = F.gen()
                     sage: f = F.monomial(2)+4*B+6
                     sage: f(1/3)
                     (66*q^2 + 66*q - 2)/(9*q^2 + 9*q)
                """
                return self.polynomial()(v)

            def polynomial(self):
                """
                Convert to a polynomial in `x`.

                EXAMPLES::

                    sage: F = QuantumValuedPolynomialRing(ZZ).S()
                    sage: S = F.gen()
                    sage: (S+1).polynomial()
                    q*x + 2

                    sage: F = QuantumValuedPolynomialRing(ZZ).B()
                    sage: B = F.gen()
                    sage: (B+1).polynomial()
                    x + 1

                TESTS::

                    sage: F.zero().polynomial().parent()
                    Univariate Polynomial Ring in x over Fraction Field
                    of Univariate Polynomial Ring in q over Integer Ring
                """
                ring = self.parent().ground_ring()
                fractions = PolynomialRing(ring, 'q').fraction_field()
                R = PolynomialRing(fractions, 'x')
                p = self.parent()._poly
                return R.sum(c * p(i) for i, c in self)

            def shift(self, j=1):
                """
                Shift all indices by `j`.

                INPUT:

                - `j` -- integer (default 1)

                In the binomial basis, the shift by 1 corresponds to
                a summation operator from `0` to `x`.

                EXAMPLES::

                    sage: F = QuantumValuedPolynomialRing(ZZ).S()
                    sage: B = F.gen()
                    sage: (B+1).shift()
                    S[1] + S[2]
                """
                A = self.parent()
                return A._from_dict({A._indices(i + j): c for i, c in self})

            def sum_of_coefficients(self):
                """
                Return the sum of coefficients.

                In the shifted basis, this is the evaluation at `x=0`.

                EXAMPLES::

                    sage: F = QuantumValuedPolynomialRing(ZZ).S()
                    sage: B = F.basis()
                    sage: (B[2]*B[4]).sum_of_coefficients()
                    1
                """
                R = self.parent().base_ring()
                return R.sum(self._monomial_coefficients.values())

    class Shifted(CombinatorialFreeModule, BindableClass):
        r"""
        The quantum-valued polynomial ring in the shifted basis.

        The basis used here is given by `S[i] = \genfrac{[}{]}{0pt}{}{i+x}{i}_q` for `i \in \NN`.

        Assuming `n_1 \leq n_2`, the product of two monomials `S[n_1] \cdot S[n_2]`
        is given by the sum

        .. MATH::

            \sum_{k=0}^{n_1} (-1)^k q^{\binom{k}{2} - n_1 * n_2} \genfrac{[}{]}{0pt}{}{n_1}{k}_q \genfrac{[}{]}{0pt}{}{n_1+n_2-k}{n_1}_q S[n_1 + n_2 - k].
        """
        def __init__(self, A):
            r"""
            Initialize ``self``.

            EXAMPLES::

                sage: F = QuantumValuedPolynomialRing(QQ).S(); F
                Quantum-Valued Polynomial Ring over Rational Field
                in the shifted basis
                sage: TestSuite(F).run()  # not tested
            """
            CombinatorialFreeModule.__init__(self, A.base_ring(),
                                             NonNegativeIntegers(),
                                             category=A.Bases(),
                                             prefix="S",
                                             latex_prefix=r"\mathbb{S}")

        def _an_element_(self):
            """
            Return a small element of ``self``.

            EXAMPLES::

                sage: F = QuantumValuedPolynomialRing(QQ).S()
                sage: F.an_element()
                2*S[0] + 4*S[2]
            """
            NonNeg = self.basis().keys()
            ring = self.base_ring()
            return self.element_class(self, {NonNeg(0): ring(2),
                                             NonNeg(2): ring(4)})

        def _realization_name(self) -> str:
            r"""
            TESTS::

                sage: F = QuantumValuedPolynomialRing(QQ).S()
                sage: F._realization_name()
                'shifted'
            """
            return "shifted"

        def product_on_basis(self, n1, n2):
            r"""
            Return the product of basis elements ``n1`` and ``n2``.

            INPUT:

            - ``n1``, ``n2`` -- integers

            EXAMPLES::

                sage: A = QuantumValuedPolynomialRing(QQ).S()
                sage: A.product_on_basis(0, 1)
                S[1]
            """
            i = ZZ(n1)
            j = ZZ(n2)
            if j < i:
                j, i = i, j
            q = self.q()
            return self._from_dict({i + j - k: (-1)**k
                                    * q_binomial(i, k)
                                    * q_binomial(i + j - k, i)
                                    * q**(binomial(k, 2) - i * j)
                                    for k in range(i + 1)})

        def _from_binomial_basis(self, i):
            """
            Convert from the ``binomial(x,k)`` basis.

            INPUT:

            - ``i`` -- an integer

            EXAMPLES::

                sage: S = QuantumValuedPolynomialRing(ZZ).S()
                sage: B = QuantumValuedPolynomialRing(ZZ).B()
                sage: b = B.basis()
                sage: S(b[3]+1)  # indirect doctest
                -(q^-6-1)*S[0] + (q^-8+q^-7+q^-6)*S[1]
                - (q^-9+q^-8+q^-7)*S[2] + (q^-9)*S[3]
                sage: B(_)
                B[0] + B[3]
            """
            i = ZZ(i)
            R = self.base_ring()
            q = self.q()
            return self._from_dict({k: R((-1)**(i - k) * q_binomial(i, k))
                                    * q**(-i**2 + binomial(i - k, 2))
                                    for k in range(i + 1)})

        def from_h_vector(self, hv):
            """
            Convert from some `h`-vector.

            EXAMPLES::

                sage: A = QuantumValuedPolynomialRing(ZZ).S()
                sage: B = A.basis()
                sage: ex = B[2] + B[3]
                sage: A.from_h_vector(ex.h_vector())
                S[2] + S[3]

                sage: q = A.base_ring().gen()
                sage: ex = B[2] + q*B[3]
                sage: A.from_h_vector(ex.h_vector())
                S[2] + q*S[3]
            """
            B = self.basis()
            ring = self.base_ring()
            q = self.q()
            d = len(hv) - 1
            m = matrix(ring, d + 1, d + 1,
                       [(-1)**(d - j) * q_binomial(d - i, d - j, q) *
                        q**(-d * (d - i) + binomial(d - j, 2))
                        for j in range(d + 1)
                        for i in range(d + 1)])
            v = vector(ring, [hv[i] for i in range(d + 1)])
            return sum(ring(c) * B[i] for i, c in enumerate(m * v))

        def _element_constructor_(self, x):
            r"""
            Convert ``x`` into ``self``.

            INPUT:

            - ``x`` -- an element of the base ring or something convertible

            EXAMPLES::

                sage: R = QuantumValuedPolynomialRing(QQ).S()
                sage: x = R.gen()
                sage: R(3)
                3*S[0]
                sage: R(x)
                S[1]
            """
            P = parent(x)
            if isinstance(P, QuantumValuedPolynomialRing.Shifted):
                if P is self:
                    return x
                if P is not self.base_ring():
                    return self.element_class(self, x.monomial_coefficients())

            # ok, not a quantum-valued polynomial ring element
            R = self.base_ring()
            # coercion via base ring
            x = R(x)
            if not x:
                return self.element_class(self, {})
            return self.from_base_ring_from_one_basis(x)

        def _coerce_map_from_(self, R) -> bool:
            r"""
            Return whether there is a coercion from ``R`` into ``self``.

            INPUT:

            - ``R`` -- a commutative ring

            The things that coerce into ``self`` are

            - Quantum-Valued Polynomial Rings over a base
              with a coercion map into ``self.base_ring()``.

            - Anything with a coercion into ``self.base_ring()``.

            EXAMPLES::

                sage: F = QuantumValuedPolynomialRing(GF(7)).S(); F
                Quantum-Valued Polynomial Ring over Finite Field of size 7
                in the shifted basis

            Elements of the quantum-valued polynomial ring canonically coerce in::

                sage: x = F.gen()
                sage: F.coerce(x*x) # indirect doctest
                (6*q^-1)*S[1] + (q^-1+1)*S[2]

            Elements of the integers coerce in, since there is a coerce map
            from `\ZZ` to GF(7)::

                sage: F.coerce(1)       # indirect doctest
                S[0]

            There is no coerce map from `\QQ` to `\GF{7}`::

                sage: F.coerce(2/3)  # indirect doctest
                Traceback (most recent call last):
                ...
                TypeError: no canonical coercion from Rational Field to
                Quantum-Valued Polynomial Ring over Finite Field of size 7
                in the shifted basis

            Elements of the base ring coerce in::

                sage: F.coerce(GF(7)(5))
                5*S[0]

            The quantum-valued polynomial ring over `\ZZ` on `x` coerces in, since
            `\ZZ` coerces to `\GF{7}`::

                sage: G = QuantumValuedPolynomialRing(ZZ).S()
                sage: Gx = G.gen()
                sage: z = F.coerce(Gx**2); z
                -(q^-1)*S[1] + (q^-1+1)*S[2]
                sage: z.parent() is F
                True

            However, `\GF{7}` does not coerce to `\ZZ`, so the shuffle
            algebra over `\GF{7}` does not coerce to the one over `\ZZ`::

                sage: G.coerce(x^3+x)
                Traceback (most recent call last):
                ...
                TypeError: no canonical coercion from Quantum-Valued Polynomial
                Ring over Finite Field of size 7 in the shifted basis
                to Quantum-Valued Polynomial Ring over Integer Ring
                in the shifted basis

            TESTS::

                sage: F = QuantumValuedPolynomialRing(ZZ).S()
                sage: G = QuantumValuedPolynomialRing(QQ).S()
                sage: H = QuantumValuedPolynomialRing(ZZ).S()
                sage: F._coerce_map_from_(G)
                False
                sage: G._coerce_map_from_(F)
                True
                sage: F._coerce_map_from_(H)
                True
                sage: F._coerce_map_from_(QQ)
                False
                sage: G._coerce_map_from_(QQ)
                True
                sage: F.has_coerce_map_from(PolynomialRing(ZZ, 'x'))
                False
            """
            # quantum-valued polynomial rings in the same variable
            # over any base that coerces in:
            if isinstance(R, QuantumValuedPolynomialRing.Shifted):
                return self.base_ring().has_coerce_map_from(R.base_ring())
            if isinstance(R, QuantumValuedPolynomialRing.Binomial):
                return R.module_morphism(self._from_binomial_basis,
                                         codomain=self)
            return self.base_ring().has_coerce_map_from(R)

        def _poly(self, i):
            """
            Convert the basis element `S[i]` to a polynomial.

            INPUT:

            - ``i`` -- an integer

            EXAMPLES::

                sage: F = QuantumValuedPolynomialRing(ZZ).S()
                sage: F._poly(4).factor()
                (1/(q^6 + 3*q^5 + 5*q^4 + 6*q^3 + 5*q^2 + 3*q + 1)) *
                (q*x + 1) * (q^2*x + q + 1) * (q^3*x + q^2 + q + 1) *
                (q^4*x + q^3 + q^2 + q + 1)
            """
            return q_binomial_x(i, i)

        class Element(CombinatorialFreeModule.Element):
            def umbra(self):
                """
                Return the Bernoulli umbra.

                This is the derivative at `-1` of the shift by one.

                This is related to Carlitz's `q`-Bernoulli numbers.

                .. SEEALSO:: :meth:`derivative_at_minus_one`

                EXAMPLES::

                    sage: F = QuantumValuedPolynomialRing(ZZ).S()
                    sage: B = F.gen()
                    sage: (B+1).umbra()
                    (q + 2)/(q + 1)
                """
                return self.shift().derivative_at_minus_one()

            def variable_shift(self, k=1):
                r"""
                Return the image by the shift on variables.

                The shift is the substitution operator

                .. MATH::

                    x \mapsto q x + 1.

                INPUT:

                - `k` -- integer (default: 1)

                EXAMPLES::

                    sage: A = QuantumValuedPolynomialRing(ZZ).S()
                    sage: S = A.basis()
                    sage: S[5].variable_shift()
                    S[0] + q*S[1] + q^2*S[2] + q^3*S[3] + q^4*S[4] + q^5*S[5]

                    sage: S[5].variable_shift(-1)
                    -(q^-5)*S[4] + (q^-5)*S[5]

                TESTS::

                    sage: S[5].variable_shift(0)
                    S[5]
                    sage: S[5].variable_shift().variable_shift(-1)
                    S[5]
                    sage: S[5].variable_shift(2).variable_shift(-2)
                    S[5]
                    sage: S[3].variable_shift(-2)
                    (q^-5)*S[1] - (q^-6+q^-5)*S[2] + (q^-6)*S[3]
                """
                if k == 0:
                    return self

                A = self.parent()
                q = A.q()

                def on_basis(n):
                    return {A._indices(j): q**(k * j)
                            * q_binomial(k + n - 1 - j, n - j)
                            for j in range(n + 1)}

                mc = self._monomial_coefficients
                ret = linear_combination((on_basis(index), coeff)
                                         for index, coeff in mc.items())
                return A.element_class(A, ret)

            def derivative_at_minus_one(self):
                """
                Return the 'derivative' at -1.

                .. SEEALSO:: :meth:`umbra`

                EXAMPLES::

                    sage: F = QuantumValuedPolynomialRing(ZZ).S()
                    sage: B = F.gen()
                    sage: (B+1).derivative_at_minus_one()
                    1
                """
                ring = q_int(1).parent()
                return ring.sum(c / q_int(i) for i, c in self if i > 0)

            def h_vector(self):
                """
                Return the numerator of the generating series of values.

                If ``self`` is an Ehrhart polynomial, this is the h-vector.

                .. SEEALSO:: :meth:`h_polynomial`, :meth:`fraction`

                EXAMPLES::

                    sage: A = QuantumValuedPolynomialRing(ZZ).S()
                    sage: ex = A.basis()[4]
                    sage: ex.h_vector()
                    (0, 0, 0, 0, 1)

                    sage: q = polygen(QQ,'q')
                    sage: x = polygen(q.parent(),'x')
                    sage: ex = A.from_polynomial((1+q*x)**3)
                    sage: ex.h_vector()
                    (0, q^3, 2*q + 2*q^2, 1)
                """
                d = max(self.support())
                ring = self.parent().base_ring()
                q = self.parent().q()

                def fn(j, i):
                    return ((-1)**(d - j) *
                            q**(binomial(d - j + i + 1, 2) -
                                binomial(i + 1, 2)) *
                            q_binomial(d - i, d - j))
                m = matrix(ring, d + 1, d + 1, fn)
                v = vector(ring, [self.coefficient(i) for i in range(d + 1)])
                return m * v

            def h_polynomial(self):
                """
                Return the `h`-vector as a polynomial.

                .. SEEALSO:: :meth:`h_vector`, :meth:`fraction`

                EXAMPLES::

                    sage: A = QuantumValuedPolynomialRing(ZZ).S()
                    sage: q = polygen(ZZ,'q')
                    sage: x = polygen(q.parent(),'x')
                    sage: ex = A.from_polynomial((1+q*x)**3)
                    sage: ex.h_polynomial()
                    z^3 + (2*q + 2*q^2)*z^2 + q^3*z
                """
                ring = PolynomialRing(self.parent().base_ring(), 'z')
                return ring(list(self.h_vector()))

            def fraction(self):
                """
                Return the generating series of values as a fraction.

                .. SEEALSO:: :meth:`h_vector`, :meth:`h_polynomial`

                EXAMPLES::

                    sage: A = QuantumValuedPolynomialRing(QQ).S()
                    sage: ex = A.basis()[4]
                    sage: ex.fraction().factor()
                    (-1) * (t - 1)^-1 * (q*t - 1)^-1 * (q^2*t - 1)^-1 * (q^3*t - 1)^-1 * (q^4*t - 1)^-1

                    sage: q = polygen(QQ,'q')
                    sage: x = polygen(q.parent(), 'x')
                    sage: ex = A.from_polynomial((1+q*x)**3)
                    sage: ex.fraction().factor()
                    (t - 1)^-1 * (q*t - 1)^-1 * (q^2*t - 1)^-1 * (q^3*t - 1)^-1 * (q^3*t^2 + 2*q^2*t + 2*q*t + 1)
                    sage: ex.fraction().numerator()
                    q^3*t^2 + 2*q^2*t + 2*q*t + 1
                """
                v = self.h_vector()
                d = len(v)
                R = PolynomialRing(QQ, 'q,t')
                frac_R = R.fraction_field()
                q, t = R.gens()
                denom = R.prod(1 - q**i * t for i in range(d))
                numer = sum(frac_R(v[i]) * t**(d - 1 - i) for i in range(d))
                return numer / denom

    S = Shifted

    # =====     Another basis for the same algebra     =====

    class Binomial(CombinatorialFreeModule, BindableClass):
        r"""
        The quantum-valued polynomial ring in the binomial basis.

        The basis used here is given by `B[i] = \genfrac{[}{]}{0pt}{}{x}{i}_q` for `i \in \NN`.

        Assuming `n_1 \leq n_2`, the product of two monomials `B[n_1] \cdot B[n_2]`
        is given by the sum

        .. MATH::

            \sum_{k=0}^{n_1} q^{(k-n_1)(k-n_2)} \genfrac{[}{]}{0pt}{}{n_1}{k}_q \genfrac{[}{]}{0pt}{}{n_1+n_2-k}{n_1}_q B[n_1 + n_2 - k].
        """
        def __init__(self, A) -> None:
            r"""
            Initialize ``self``.

            EXAMPLES::

                sage: F = QuantumValuedPolynomialRing(QQ).B(); F
                Quantum-Valued Polynomial Ring over Rational Field
                in the binomial basis
                sage: TestSuite(F).run()  # not tested
            """
            CombinatorialFreeModule.__init__(self, A.base_ring(),
                                             NonNegativeIntegers(),
                                             category=A.Bases(),
                                             prefix="B",
                                             latex_prefix=r"\mathbb{B}")

        def _realization_name(self) -> str:
            r"""
            TESTS::

                sage: F = QuantumValuedPolynomialRing(QQ).B()
                sage: F._realization_name()
                'binomial'
            """
            return "binomial"

        def product_on_basis(self, n1, n2):
            r"""
            Return the product of basis elements ``n1`` and ``n2``.

            INPUT:

            - ``n1``, ``n2`` -- integers

            The formula is taken from Theorem 3.4 in [HaHo2017]_.

            EXAMPLES::

                sage: A = QuantumValuedPolynomialRing(QQ).B()
                sage: A.product_on_basis(0, 1)
                B[1]
            """
            i = ZZ(n1)
            j = ZZ(n2)
            if j < i:
                j, i = i, j

            q = self.q()
            return self._from_dict({i + j - k:
                                    q_binomial(i, k)
                                    * q_binomial(i + j - k, i)
                                    * q**((k - i) * (k - j))
                                    for k in range(i + 1)})

        def _from_shifted_basis(self, i):
            """
            Convert from the shifted binomial(x+k,k) basis.

            INPUT:

            - ``i`` -- an integer

            EXAMPLES::

                sage: S = QuantumValuedPolynomialRing(ZZ).S()
                sage: B = QuantumValuedPolynomialRing(ZZ).B()
                sage: s = S.basis()
                sage: B(s[3]+1)  # indirect doctest
                2*B[0] + (q+q^2+q^3)*B[1] + (q^4+q^5+q^6)*B[2] + q^9*B[3]
                sage: S(_)
                S[0] + S[3]
            """
            i = ZZ(i)
            R = self.base_ring()
            q = self.q()
            return self._from_dict({k: R(q_binomial(i, k))
                                    * q**(k**2)
                                    for k in range(i + 1)})

        def _element_constructor_(self, x):
            r"""
            Convert ``x`` into ``self``.

            EXAMPLES::

                sage: R = QuantumValuedPolynomialRing(QQ).B()
                sage: x = R.gen()
                sage: R(3)
                3*B[0]
                sage: R(x)
                B[1]
            """
            P = x.parent()
            if isinstance(P, QuantumValuedPolynomialRing.Binomial):
                if P is self:
                    return x
                if P is not self.base_ring():
                    return self.element_class(self, x.monomial_coefficients())

            # ok, not a quantum-valued polynomial ring element
            R = self.base_ring()
            # coercion via base ring
            x = R(x)
            if not x:
                return self.element_class(self, {})
            return self.from_base_ring_from_one_basis(x)

        def _coerce_map_from_(self, R):
            r"""
            Return whether there is a coercion from ``R`` into ``self``.

            INPUT:

            - ``R`` -- a commutative ring

            The things that coerce into ``self`` are

            - Quantum-Valued Polynomial Rings over a base
              with a coercion map into ``self.base_ring()``.

            - Anything with a coercion into ``self.base_ring()``.

            EXAMPLES::

                sage: F = QuantumValuedPolynomialRing(GF(7)).B(); F
                Quantum-Valued Polynomial Ring over Finite Field of size 7
                in the binomial basis

            Elements of the integer-valued polynomial ring canonically coerce
            in::

                sage: x = F.gen()
                sage: F.coerce(x*x) # indirect doctest
                B[1] + (q+q^2)*B[2]

            Elements of the integers coerce in, since there is a coerce map
            from `\ZZ` to `\GF(7)`::

                sage: F.coerce(1)       # indirect doctest
                B[0]

            There is no coerce map from `\QQ` to `\GF{7}`::

                sage: F.coerce(2/3)  # indirect doctest
                Traceback (most recent call last):
                ...
                TypeError: no canonical coercion from Rational Field to
                Quantum-Valued Polynomial Ring over Finite Field of size 7
                in the binomial basis

            Elements of the base ring coerce in::

                sage: F.coerce(GF(7)(5))
                5*B[0]

            The integer-valued polynomial ring over `\ZZ` on `x` coerces in,
            since `\ZZ` coerces to `\GF{7}`::

                sage: G = QuantumValuedPolynomialRing(ZZ).B()
                sage: Gx = G.gen()
                sage: z = F.coerce(Gx**2); z
                B[1] + (q+q^2)*B[2]
                sage: z.parent() is F
                True

            However, `\GF{7}` does not coerce to `\ZZ`, so the
            integer-valued polynomial algebra over `\GF{7}` does not
            coerce to the one over `\ZZ`::

                sage: G.coerce(x^3+x)
                Traceback (most recent call last):
                ...
                TypeError: no canonical coercion from
                Quantum-Valued Polynomial Ring over Finite Field of size 7
                in the binomial basis to Quantum-Valued Polynomial Ring
                over Integer Ring in the binomial basis

            TESTS::

                sage: F = QuantumValuedPolynomialRing(ZZ).B()
                sage: G = QuantumValuedPolynomialRing(QQ).B()
                sage: H = QuantumValuedPolynomialRing(ZZ).B()
                sage: F._coerce_map_from_(G)
                False
                sage: G._coerce_map_from_(F)
                True
                sage: F._coerce_map_from_(H)
                True
                sage: F._coerce_map_from_(QQ)
                False
                sage: G._coerce_map_from_(QQ)
                True
                sage: F.has_coerce_map_from(PolynomialRing(ZZ,'x'))
                False
            """
            # quantum-valued polynomial rings over any base
            # that coerces in:
            if isinstance(R, QuantumValuedPolynomialRing.Binomial):
                return self.base_ring().has_coerce_map_from(R.base_ring())
            if isinstance(R, QuantumValuedPolynomialRing.Shifted):
                return R.module_morphism(self._from_shifted_basis,
                                         codomain=self)
            return self.base_ring().has_coerce_map_from(R)

        def _poly(self, i):
            """
            Convert the basis element `B[i]` to a polynomial.

            INPUT:

            - ``i`` -- an integer

            EXAMPLES::

                sage: F = QuantumValuedPolynomialRing(ZZ).B()
                sage: F._poly(4).factor()
                (1/(q^12 + 3*q^11 + 5*q^10 + 6*q^9 + 5*q^8 + 3*q^7 + q^6)) *
                (x - 1) * x * (x - q - 1) * (x - q^2 - q - 1)
            """
            return q_binomial_x(0, i)

        class Element(CombinatorialFreeModule.Element):
            def variable_shift(self, k=1):
                r"""
                Return the image by the shift of variables.

                On polynomials, the action for `k=1` is the shift
                on variables `x \mapsto 1 + qx`.

                This implementation follows formula (5.5) in [HaHo2017]_.

                INPUT:

                - `k` -- nonnegative integer (default: 1)

                EXAMPLES::

                    sage: A = QuantumValuedPolynomialRing(ZZ).B()
                    sage: B = A.basis()
                    sage: B[5].variable_shift()
                    B[4] + q^5*B[5]

                TESTS::

                    sage: B[5].variable_shift(0)
                    B[5]
                """
                if k == 0:
                    return self

                A = self.parent()
                q = A.q()

                def on_basis(n):
                    return {A._indices(j): q**((k + j - n) * j)
                            * q_binomial(k, n - j)
                            for j in range(n + 1)}

                mc = self._monomial_coefficients
                ret = linear_combination((on_basis(index), coeff)
                                         for index, coeff in mc.items())
                return A.element_class(A, ret)

    B = Binomial



# distutils: libraries = NTL_LIBRARIES gmp m
# distutils: extra_compile_args = NTL_CFLAGS
# distutils: include_dirs = NTL_INCDIR
# distutils: library_dirs = NTL_LIBDIR
# distutils: extra_link_args = NTL_LIBEXTRA
# distutils: language = c++

# ****************************************************************************
#       Copyright (C) 2005 William Stein <wstein@gmail.com>
#       Copyright (C) 2007 Martin Albrecht <malb@informatik.uni-bremen.de>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.ext.cplusplus cimport ccrepr, ccreadstr

include 'misc.pxi'
include 'decl.pxi'

from cpython.object cimport Py_EQ, Py_NE
from sage.rings.integer cimport Integer

from sage.libs.ntl.ntl_ZZ import unpickle_class_value
from sage.libs.ntl.ntl_GF2 cimport ntl_GF2


##############################################################################
#
# ntl_GF2X: Polynomials over GF(2) via NTL
#
# AUTHORS:
#  - Martin Albrecht <malb@informatik.uni-bremen.de>
#    2006-01: initial version (based on code by William Stein)
#  - Martin Albrecht <malb@informatik.uni-bremen.de>
#    2007-10: adapted to new conventions
#
##############################################################################

def GF2XHexOutput(have_hex=None):
    """
    Represent GF2X and GF2E elements in the more compact
    hexadecimal form to the user.

    If no parameter is provided the currently set value will be
    returned.

    INPUT:

    - ``have_hex`` -- if ``True`` hex representation will be used

    EXAMPLES::

        sage: m = ntl.GF2EContext(ntl.GF2X([1,1,0,1,1,0,0,0,1]))
        sage: x = ntl.GF2E([1,0,1,0,1], m) ; x
        [1 0 1 0 1]

        sage: ntl.GF2XHexOutput() ## indirect doctest
        False
        sage: ntl.GF2XHexOutput(True)
        sage: ntl.GF2XHexOutput()
        True

        sage: x
        0x51

        sage: ntl.GF2XHexOutput(False)
        sage: x
        [1 0 1 0 1]
    """
    if have_hex is None:
        return bool(GF2XHexOutput_c[0])

    if have_hex:
        GF2XHexOutput_c[0] = 1
    else:
        GF2XHexOutput_c[0] = 0


cdef class ntl_GF2X():
    """
    Univariate Polynomials over GF(2) via NTL.
    """
    def __init__(self, x=[]):
        """
        Construct a new polynomial over GF(2).

        A value may be passed to this constructor. If you pass a string
        to the constructor please note that byte sequences and the hexadecimal
        notation are little endian.  So e.g. '[0 1]' == '0x2' == x.

        Input types are ntl.ZZ_px, strings, lists of digits, FiniteFieldElements
        from extension fields over GF(2), Polynomials over GF(2), Integers, and finite
        extension fields over GF(2) (uses modulus).

        INPUT:

        - ``x`` -- value to be assigned to this element. See examples.

        OUTPUT: a new ntl.GF2X element

        EXAMPLES::

            sage: ntl.GF2X(ntl.ZZ_pX([1,1,3],2))
            [1 1 1]
            sage: ntl.GF2X('0x1c')
            [1 0 0 0 0 0 1 1]
            sage: ntl.GF2X('[1 0 1 0]')
            [1 0 1]
            sage: ntl.GF2X([1,0,1,0])
            [1 0 1]
            sage: ntl.GF2X(GF(2**8,'a').gen()**20)
            [0 0 1 0 1 1 0 1]
            sage: ntl.GF2X(GF(2**8,'a'))
            [1 0 1 1 1 0 0 0 1]
            sage: ntl.GF2X(2)
            [0 1]
            sage: ntl.GF2X(ntl.GF2(1))
            [1]

            sage: R.<x> = GF(2)[]
            sage: f = x^5+x^2+1
            sage: ntl.GF2X(f)
            [1 0 1 0 0 1]
        """
        from sage.rings.finite_rings.element_givaro import FiniteField_givaroElement
        from sage.rings.finite_rings.element_ntl_gf2e import FiniteField_ntl_gf2eElement
        from sage.rings.finite_rings.finite_field_base import FiniteField
        from sage.rings.polynomial.polynomial_gf2x import Polynomial_GF2X

        cdef long _x

        if isinstance(x, ntl_GF2):
            GF2X_conv_GF2(self.x,(<ntl_GF2>x).x)
            return
        elif isinstance(x, ntl_GF2X):
            self.x = (<ntl_GF2X>x).x
            return
        elif isinstance(x, int):
            _x = x
            GF2XFromBytes(self.x, <unsigned char *>(&_x),sizeof(long))
            return

        if isinstance(x, Integer):
            #binary repr, reversed, and "["..."]" added
            x="["+x.binary()[::-1].replace(""," ")+"]"
        elif isinstance(x, Polynomial_GF2X):
            x = x.list() # this is slow but cimport leads to circular imports
        elif isinstance(x, FiniteField):
            if x.characteristic() == 2:
                x = list(x.modulus())
        elif isinstance(x, FiniteField_givaroElement):
            x = "0x" + hex(x.to_integer())[2:][::-1]
        elif isinstance(x, FiniteField_ntl_gf2eElement):
            x = x.polynomial().list()
        s = str(x).replace(","," ")
        # TODO: this is very slow, but we wait until somebody complains
        ccreadstr(self.x, s)

    def __reduce__(self):
        """
        EXAMPLES::

            sage: f = ntl.GF2X(ntl.ZZ_pX([1,1,3],2))
            sage: loads(dumps(f)) == f
            True
            sage: f = ntl.GF2X('0x1c')
            sage: loads(dumps(f)) == f
            True
        """
        return unpickle_class_value, (ntl_GF2X, self.hex())

    def __repr__(self):
        """
        Return the string representation of ``self``.

        EXAMPLES::

            sage: ntl.GF2X(ntl.ZZ_pX([1,1,3],2)).__repr__()
            '[1 1 1]'
        """
        return ccrepr(self.x)

    def __mul__(ntl_GF2X self, other):
        """
        EXAMPLES::

            sage: f = ntl.GF2X([1,0,1,1]) ; g = ntl.GF2X([0,1])
            sage: f*g ## indirect doctest
            [0 1 0 1 1]
        """
        cdef ntl_GF2X r = ntl_GF2X.__new__(ntl_GF2X)
        if not isinstance(other, ntl_GF2X):
            other = ntl_GF2X(other)
        GF2X_mul(r.x, self.x, (<ntl_GF2X>other).x)
        return r

    def __truediv__(ntl_GF2X self, b):
        """
        EXAMPLES::

            sage: a = ntl.GF2X(4)
            sage: a / ntl.GF2X(2)
            [0 1]
            sage: a / ntl.GF2X(3)
            Traceback (most recent call last):
            ...
            ArithmeticError: self (=[0 0 1]) is not divisible by b (=[1 1])
        """
        cdef ntl_GF2X q = ntl_GF2X.__new__(ntl_GF2X)
        cdef int divisible

        if not isinstance(b, ntl_GF2X):
            b = ntl_GF2X(b)

        divisible = GF2X_divide(q.x, self.x, (<ntl_GF2X>b).x)
        if not divisible:
            raise ArithmeticError("self (=%s) is not divisible by b (=%s)" % (self, b))
        return q

    def DivRem(ntl_GF2X self, b):
        """
        EXAMPLES::

            sage: a = ntl.GF2X(4)
            sage: a.DivRem( ntl.GF2X(2) )
            ([0 1], [])
            sage: a.DivRem( ntl.GF2X(3) )
            ([1 1], [1])
        """
        cdef ntl_GF2X q = ntl_GF2X.__new__(ntl_GF2X)
        cdef ntl_GF2X r = ntl_GF2X.__new__(ntl_GF2X)

        if not isinstance(b, ntl_GF2X):
            b = ntl_GF2X(b)

        GF2X_DivRem(q.x, r.x, self.x, (<ntl_GF2X>b).x)
        return q,r

    def __floordiv__(ntl_GF2X self, b):
        """
        EXAMPLES::

            sage: a = ntl.GF2X(4)
            sage: a // ntl.GF2X(2)
            [0 1]
            sage: a // ntl.GF2X(3)
            [1 1]
        """
        cdef ntl_GF2X q = ntl_GF2X.__new__(ntl_GF2X)

        if not isinstance(b, ntl_GF2X):
            b = ntl_GF2X(b)

        GF2X_div(q.x, self.x, (<ntl_GF2X>b).x)
        return q

    def __mod__(ntl_GF2X self, b):
        """
        EXAMPLES::

            sage: a = ntl.GF2X(4)
            sage: a % ntl.GF2X(2)
            []
            sage: a % ntl.GF2X(3)
            [1]
        """
        cdef ntl_GF2X r = ntl_GF2X.__new__(ntl_GF2X)

        if not isinstance(b, ntl_GF2X):
            b = ntl_GF2X(b)

        GF2X_rem(r.x, self.x, (<ntl_GF2X>b).x)
        return r

    def __sub__(ntl_GF2X self, other):
        """
        EXAMPLES::

            sage: f = ntl.GF2X([1,0,1,1]) ; g = ntl.GF2X([0,1])
            sage: f - g ## indirect doctest
            [1 1 1 1]
            sage: g - f
            [1 1 1 1]
        """
        cdef ntl_GF2X r = ntl_GF2X.__new__(ntl_GF2X)
        if not isinstance(other, ntl_GF2X):
            other = ntl_GF2X(other)
        GF2X_sub(r.x, self.x, (<ntl_GF2X>other).x)
        return r

    def __add__(ntl_GF2X self, other):
        """
        EXAMPLES::

            sage: f = ntl.GF2X([1,0,1,1]) ; g = ntl.GF2X([0,1,0])
            sage: f + g ## indirect doctest
            [1 1 1 1]
        """
        cdef ntl_GF2X r = ntl_GF2X.__new__(ntl_GF2X)
        if not isinstance(other, ntl_GF2X):
            other = ntl_GF2X(other)
        GF2X_add(r.x, self.x, (<ntl_GF2X>other).x)
        return r

    def __neg__(ntl_GF2X self):
        """
        EXAMPLES::

            sage: f = ntl.GF2X([1,0,1,1])
            sage: -f ## indirect doctest
            [1 0 1 1]
            sage: f == -f
            True
        """
        cdef ntl_GF2X r = ntl_GF2X.__new__(ntl_GF2X)
        GF2X_negate(r.x, self.x)
        return r

    def __pow__(ntl_GF2X self, long e, ignored):
        """
        EXAMPLES::

            sage: f = ntl.GF2X([1,0,1,1]) ; g = ntl.GF2X([0,1,0])
            sage: f**3 ## indirect doctest
            [1 0 1 1 1 0 0 1 1 1]
        """
        cdef ntl_GF2X r = ntl_GF2X.__new__(ntl_GF2X)
        GF2X_power(r.x, self.x, e)
        return r

    def __richcmp__(ntl_GF2X self, other, int op):
        """
        Compare ``self`` to ``other``.

        EXAMPLES::

            sage: f = ntl.GF2X([1,0,1,1])
            sage: g = ntl.GF2X([0,1,0])
            sage: f == g ## indirect doctest
            False
            sage: f == f
            True
            sage: g != polygen(GF(2))
            False
        """
        if op != Py_EQ and op != Py_NE:
            raise TypeError("elements of GF(2)[X] are not ordered")

        cdef ntl_GF2X b
        try:
            b = <ntl_GF2X?>other
        except TypeError:
            b = ntl_GF2X(other)

        return (op == Py_EQ) == (self.x == b.x)

    def __lshift__(ntl_GF2X self, int i):
        """
        Return left shift of ``self`` by i bits ( == multiplication by
        `X^i`).

        INPUT:

        - ``i`` -- offset/power of X

        EXAMPLES::

            sage: a = ntl.GF2X(4); a
            [0 0 1]
            sage: a << 2
            [0 0 0 0 1]
        """
        cdef ntl_GF2X r = ntl_GF2X.__new__(ntl_GF2X)
        GF2X_LeftShift(r.x, self.x, <long>i)
        return r

    def __rshift__(ntl_GF2X self, int offset):
        """
        Return right shift of ``self`` by i bits ( == floor division by
        `X^i`).

        INPUT:

        - ``i`` -- offset/power of X

        EXAMPLES::

            sage: a = ntl.GF2X(4); a
            [0 0 1]
            sage: a >> 1
            [0 1]
        """
        cdef ntl_GF2X r = ntl_GF2X.__new__(ntl_GF2X)
        GF2X_RightShift(r.x, self.x, <long>offset)
        return r

    def GCD(ntl_GF2X self, other):
        """
        Return GCD of ``self`` and ``other``.

        INPUT:

        - ``other`` -- ntl.GF2X

        EXAMPLES::

            sage: a = ntl.GF2X(10)
            sage: b = ntl.GF2X(4)
            sage: a.GCD(b)
            [0 1]
        """
        cdef ntl_GF2X gcd = ntl_GF2X.__new__(ntl_GF2X)

        if not isinstance(other, ntl_GF2X):
            other = ntl_GF2X(other)

        GF2X_GCD(gcd.x, self.x, (<ntl_GF2X>other).x)
        return gcd

    def XGCD(ntl_GF2X self, other):
        """
        Return the extended gcd of ``self`` and ``other``, i.e., elements r, s, t such that.

            r = s  * self + t  * other.

        INPUT:

        - ``other`` -- ntl.GF2X

        EXAMPLES::

            sage: a = ntl.GF2X(10)
            sage: b = ntl.GF2X(4)
            sage: r,s,t = a.XGCD(b)
            sage: r == a*s + t*b
            True
        """
        cdef ntl_GF2X r = ntl_GF2X.__new__(ntl_GF2X)
        cdef ntl_GF2X s = ntl_GF2X.__new__(ntl_GF2X)
        cdef ntl_GF2X t = ntl_GF2X.__new__(ntl_GF2X)

        if not isinstance(other, ntl_GF2X):
            other = ntl_GF2X(other)

        GF2X_XGCD(r.x, s.x, t.x, self.x, (<ntl_GF2X>other).x)
        return r,s,t

    def deg(ntl_GF2X self):
        """
        Return the degree of this polynomial.

        EXAMPLES::

            sage: ntl.GF2X([1,0,1,1]).deg()
            3
        """
        return GF2X_deg(self.x)

    def list(ntl_GF2X self):
        """
        Represent this element as a list of binary digits.

        EXAMPLES::

             sage: e=ntl.GF2X([0,1,1])
             sage: e.list()
             [0, 1, 1]
             sage: e=ntl.GF2X('0xff')
             sage: e.list()
             [1, 1, 1, 1, 1, 1, 1, 1]

        OUTPUT: list of digits representing the coefficients in this element's
        polynomial representation
        """
        return [self[i] for i in range(GF2X_deg(self.x)+1)]

    def bin(ntl_GF2X self):
        r"""
        Return binary representation of this element.

        It is the same as setting \code{ntl.GF2XHexOutput(False)} and
        representing this element afterwards. However it should be
        faster and preserves the HexOutput state as opposed to the above code.

        EXAMPLES::

             sage: e=ntl.GF2X([1,1,0,1,1,1,0,0,1])
             sage: e.bin()
             '[1 1 0 1 1 1 0 0 1]'

        OUTPUT:

        string representing this element in binary digits
        """
        cdef long _hex = GF2XHexOutput_c[0]
        GF2XHexOutput_c[0] = 0
        s = ccrepr(self.x)
        GF2XHexOutput_c[0] = _hex
        return s

    def hex(ntl_GF2X self):
        r"""
        Return a hexadecimal representation of this element.

        It is the same as setting \code{ntl.GF2XHexOutput(True)} and
        representing this element afterwards. However it should be faster and
        preserves the HexOutput state as opposed to the above code.

        OUTPUT: string representing this element in hexadecimal

        EXAMPLES::

            sage: e = ntl.GF2X([1,1,0,1,1,1,0,0,1])
            sage: e.hex()
            '0xb31'
        """
        cdef long _hex = GF2XHexOutput_c[0]
        GF2XHexOutput_c[0] = 1
        s = ccrepr(self.x)
        GF2XHexOutput_c[0] = _hex
        return s

    def __hash__(self):
        return hash(self.hex())

    def _sage_(ntl_GF2X self, R=None):
        """
        Return a Sage polynomial over GF(2) equivalent to this element.

        If a ring R is provided, it is used to construct the polynomial
        in. Otherwise, an appropriate ring is generated.

        INPUT:

        - ``self`` -- GF2X element
        - ``R`` -- PolynomialRing over GF(2)

        OUTPUT: polynomial in R

        EXAMPLES::

            sage: f = ntl.GF2X([1,0,1,1,0,1])
            sage: f._sage_()
            x^5 + x^3 + x^2 + 1
            sage: f._sage_(PolynomialRing(Integers(2),'y'))
            y^5 + y^3 + y^2 + 1
        """
        if R is None:
            from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
            from sage.rings.finite_rings.finite_field_constructor import FiniteField
            R = PolynomialRing(FiniteField(2), 'x')

        return R([int(c) for c in self.list()])

    def coeff(self, int i):
        """
        Return the coefficient of the monomial `X^i` in ``self``.

        INPUT:

        - ``i`` -- degree of X

        EXAMPLES::

            sage: e = ntl.GF2X([0,1,0,1])
            sage: e.coeff(0)
            0
            sage: e.coeff(1)
            1
        """
        cdef ntl_GF2 c = ntl_GF2.__new__(ntl_GF2)
        c.x = GF2X_coeff(self.x, i)
        return c

    def __getitem__(self, int i):
        """
        EXAMPLES::

            sage: e = ntl.GF2X([0,1,0,1])
            sage: e[0] # indirect doctest
            0
            sage: e[1]
            1
        """
        cdef ntl_GF2 c = ntl_GF2.__new__(ntl_GF2)
        c.x = GF2X_coeff(self.x, i)
        return c

    def LeadCoeff(self):
        """
        Return the leading coefficient of ``self``.

        This is always 1 except when ``self == 0``.

        EXAMPLES::

            sage: e = ntl.GF2X([0,1])
            sage: e.LeadCoeff()
            1
            sage: e = ntl.GF2X(0)
            sage: e.LeadCoeff()
            0
        """
        cdef ntl_GF2 c = ntl_GF2.__new__(ntl_GF2)
        c.x = GF2X_LeadCoeff(self.x)
        return c

    def ConstTerm(self):
        """
        Return the constant term of ``self``.

        EXAMPLES::

            sage: e = ntl.GF2X([1,0,1])
            sage: e.ConstTerm()
            1
            sage: e = ntl.GF2X(0)
            sage: e.ConstTerm()
            0
        """
        cdef ntl_GF2 c = ntl_GF2.__new__(ntl_GF2)
        c.x = GF2X_ConstTerm (self.x)
        return c

    def SetCoeff(self, int i, a):
        """
        Set the value of a coefficient of ``self``.

        EXAMPLES::

            sage: e = ntl.GF2X([1,0,1]); e
            [1 0 1]
            sage: e.SetCoeff(1,1)
            sage: e
            [1 1 1]
        """
        cdef ntl_GF2 _a = ntl_GF2(a)

        GF2X_SetCoeff(self.x, i, _a.x)

    def __setitem__(self, int i, a):
        """
        EXAMPLES::

            sage: e = ntl.GF2X([1,0,1]); e
            [1 0 1]
            sage: e[1] = 1 # indirect doctest
            sage: e
            [1 1 1]
        """
        cdef ntl_GF2 _a = ntl_GF2(a)
        GF2X_SetCoeff(self.x, i, _a.x)

    def diff(self):
        """
        Differentiate ``self``.

        EXAMPLES::

            sage: e = ntl.GF2X([1,0,1,1,0])
            sage: e.diff()
            [0 0 1]
        """
        cdef ntl_GF2X d = ntl_GF2X.__new__(ntl_GF2X)
        d.x = GF2X_diff(self.x)
        return d

    def reverse(self, int hi=-2):
        """
        Return reverse of a[0]..a[hi] (hi >= -1)
        hi defaults to deg(a)

        INPUT:

        - ``hi`` -- bit position until which reverse is requested

        EXAMPLES::

            sage: e = ntl.GF2X([1,0,1,1,0])
            sage: e.reverse()
            [1 1 0 1]
        """
        cdef ntl_GF2X r = ntl_GF2X.__new__(ntl_GF2X)
        if hi < -1:
            hi = GF2X_deg(self.x)
        r.x = GF2X_reverse(self.x, hi)
        return r

    def weight(self):
        """
        Return the number of nonzero coefficients in ``self``.

        EXAMPLES::

            sage: e = ntl.GF2X([1,0,1,1,0])
            sage: e.weight()
            3
        """
        return int(GF2X_weight(self.x))

    def __int__(self):
        """
        EXAMPLES::

            sage: e = ntl.GF2X([1,0,1,1,0])
            sage: int(e)
            Traceback (most recent call last):
            ...
            ValueError: cannot convert non-constant polynomial to integer
            sage: e = ntl.GF2X([1])
            sage: int(e)
            1
        """
        if GF2X_deg(self.x) != 0:
            raise ValueError("cannot convert non-constant polynomial to integer")
        else:
            return GF2_conv_to_long(GF2X_coeff(self.x,0))

    def NumBits(self):
        """
        Return the number of bits of self, i.e., deg(self) + 1.

        EXAMPLES::

            sage: e = ntl.GF2X([1,0,1,1,0])
            sage: e.NumBits()
            4
        """
        return int(GF2X_NumBits(self.x))

    def __len__(self):
        """
        EXAMPLES::

            sage: e = ntl.GF2X([1,0,1,1,0])
            sage: len(e)
            4
        """
        return int(GF2X_NumBits(self.x))

    def NumBytes(self):
        """
        Return the number of bytes of ``self``, i.e., floor((NumBits(self)+7)/8).

        EXAMPLES::

            sage: e = ntl.GF2X([1,0,1,1,0,0,0,0,1,1,1,0,0,1,1,0,1,1])
            sage: e.NumBytes()
            3
        """
        return int(GF2X_NumBytes(self.x))



###########################################################
# Re-bindings for unpickling
#
# We want to ensure class
# sage.modular.congroup_element.CongruenceSubgroupElement still exists, so we
# can unpickle safely.
#
###########################################################

from sage.modular.arithgroup.arithgroup_element import ArithmeticSubgroupElement

CongruenceSubgroupElement = ArithmeticSubgroupElement



r"""
Transversal matroids

A transversal matroid arises from a groundset `E` and a collection `A` of sets
over the groundset. This can be modeled as a bipartite graph `B`, where the
vertices on the left are groundset elements, the vertices on the right are the
sets, and edges represent containment. Then a set `X` from the groundset is
independent if and only if `X` has a matching in `B`.

To construct a transversal matroid, first import
:class:`~sage.matroids.transversal_matroid.TransversalMatroid` from
:mod:`sage.matroids.transversal_matroid`.
The input should be a set system, formatted as an iterable of iterables::

    sage: from sage.matroids.transversal_matroid import TransversalMatroid
    sage: sets = [[3, 4, 5, 6, 7, 8]] * 3
    sage: M = TransversalMatroid(sets); M
    Transversal matroid of rank 3 on 6 elements, with 3 sets
    sage: M.groundset()
    frozenset({3, 4, 5, 6, 7, 8})
    sage: M.is_isomorphic(matroids.Uniform(3, 6))
    True
    sage: M = TransversalMatroid([[0, 1], [1, 2, 3], [3, 4, 5]],
    ....:                        set_labels=['1', '2', '3'])
    sage: M.graph().vertices()
    ['1', '2', '3', 0, 1, 2, 3, 4, 5]

AUTHORS:

- Zachary Gershkoff (2017-08-07): initial version

REFERENCES:

- [Bon2017]_
"""

# *****************************************************************************
#       Copyright (C) 2017 Zachary Gershkoff <zgersh2@lsu.edu>
#
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# *****************************************************************************

from collections import Counter
from copy import copy
from cpython.object cimport Py_EQ, Py_NE

from sage.graphs.digraph import DiGraph
from sage.graphs.bipartite_graph import BipartiteGraph
from sage.matroids.basis_exchange_matroid cimport BasisExchangeMatroid
from sage.matroids.minor_matroid import MinorMatroid
from sage.matroids.utilities import newlabel


cdef class TransversalMatroid(BasisExchangeMatroid):
    r"""
    The Transversal Matroid class.

    INPUT:

    - ``sets`` -- iterable of iterables of elements
    - ``groundset`` -- (optional) iterable containing names of groundset
      elements
    - ``set_labels`` -- (optional) list of labels in 1-1 correspondence with
      the iterables in ``sets``
    - ``matching`` -- (optional) dictionary specifying a maximal matching
      between elements and set labels

    OUTPUT:

    An instance of ``TransversalMatroid``. The sets specified in ``sets``
    define the matroid. If ``matching`` is not specified, the constructor
    will determine a matching to use for basis exchange.

    EXAMPLES::

        sage: from sage.matroids.transversal_matroid import TransversalMatroid
        sage: sets = [[0, 1, 2, 3]] * 3
        sage: M = TransversalMatroid(sets)
        sage: M.full_rank()
        3
        sage: M.bases_count()
        4
        sage: sum(1 for b in M.bases())
        4

    ::

        sage: from sage.matroids.transversal_matroid import TransversalMatroid
        sage: M = TransversalMatroid(sets=[['a', 'c']], groundset=['a', 'c', 'd'])
        sage: M.loops()
        frozenset({'d'})
        sage: M.full_rank()
        1
    """

    def __init__(self, sets, groundset=None, set_labels=None, matching=None):
        """
        See the class definition for full documentation.

        EXAMPLES::

            sage: from sage.matroids.transversal_matroid import TransversalMatroid
            sage: sets = [[0, 1, 2, 3], [1, 2], [1, 3, 4]]
            sage: set_labels = [5, 6, 7]
            sage: M = TransversalMatroid(sets, set_labels=set_labels)
            sage: TestSuite(M).run()

        TESTS::

            sage: from sage.matroids.transversal_matroid import TransversalMatroid
            sage: M = TransversalMatroid([[0, 1], [1, 2], [2, 3], [3, 4]])
            sage: TestSuite(M).run()

            sage: M = TransversalMatroid([[], [], []], groundset=range(3)); M
            Transversal matroid of rank 0 on 3 elements, with 3 sets
            sage: sets = [[0, 1, 2, 3, 4], [4, 5]]
            sage: M = TransversalMatroid(sets, groundset=[0, 1, 2, 3])
            Traceback (most recent call last):
            ...
            ValueError: groundset and sets do not match
            sage: M = TransversalMatroid(sets, set_labels=[1, 2, 3])
            Traceback (most recent call last):
            ...
            ValueError: set labels do not match sets
            sage: M = TransversalMatroid(sets, set_labels=[1, 2])
            Traceback (most recent call last):
            ...
            ValueError: set labels cannot be element labels

            sage: sets = [['s1', 's2'], ['s1', 's3']]
            sage: M = TransversalMatroid(sets); M
            Transversal matroid of rank 2 on 3 elements, with 2 sets
            sage: M.sets()
            [['s1', 's2'], ['s1', 's3']]
            sage: M.set_labels()
            ['s0', 's4']

        Testing the matching parameter::

            sage: M = TransversalMatroid([range(5)] * 4, set_labels='abcd')
            sage: M.full_rank()
            4
            sage: M.rank([0, 1, 2])
            3

        ::

            sage: sets = [[0, 1, 2, 3], [1, 2], [1, 3, 4]]
            sage: set_labels = [5, 6, 7]
            sage: m = {0: 5, 1: 6, 3: 7}
            sage: M = TransversalMatroid(sets, set_labels=set_labels, matching=m)
            sage: len(M.bases())
            9
        """
        contents = {e for subset in sets for e in subset}
        if groundset is None:
            groundset = contents
        elif not contents.issubset(groundset):
            raise ValueError("groundset and sets do not match")
        groundset = tuple(groundset)

        # keep the original list as input so we don't lose order between minors etc.
        self._sets_input = [list(s) for s in sets]
        self._sets = Counter([frozenset(s) for s in self._sets_input])

        # This might be redundant with self._idx from BasisExchangeMatroid
        # However, BasisExchangeMatroid has not been called yet
        element_int_map = {e: i for i, e in enumerate(groundset)}
        int_element_map = {i: e for i, e in enumerate(groundset)}

        # we need a matching and a corresponding graph
        if set_labels:
            if len(set_labels) != len(sets):
                raise ValueError("set labels do not match sets")
            if not contents.isdisjoint(set_labels):
                raise ValueError("set labels cannot be element labels")
            if matching:
                matching_temp = matching
            self._set_labels_input = list(set_labels)

        else:
            if matching:
                raise ValueError("set labels must be provided if matching is provided")
            self._set_labels_input = []
            increment = 0
            for i in range(len(sets)):  # make sure "sn" is not in groundset
                label = 's' + str(i + increment)
                while label in groundset:
                    increment += 1
                    label = 's' + str(i + increment)
                self._set_labels_input.append(label)
        # for internal use with integers representing elements
        set_labels = ['s' + str(i) for i in range(len(sets))]

        if not matching:
            B = BipartiteGraph()
            for e in groundset:
                B.add_vertex(element_int_map[e], left=True)
            for i, s in enumerate(sets):
                new_vertex = set_labels[i]
                for e in s:
                    B.add_edge(new_vertex, element_int_map[e])
            matching_temp = {}
            for u, v, _ in B.matching():
                if u in range(len(groundset)):
                    matching_temp[int_element_map[u]] = v
                else:
                    matching_temp[int_element_map[v]] = u

        self._set_labels = list(set_labels)

        # determine the basis from the matching
        basis = frozenset(matching_temp)

        # This creates self._groundset attribute, among other things
        # It takes the actual groundset labels, not the translated ones
        BasisExchangeMatroid.__init__(self, groundset, basis)

        # matching_temp uses actual groundset labels
        # self._matching will use the translated ones
        if matching:
            self._matching = {self._idx[e]: self._set_labels[self._set_labels_input.index(matching_temp[e])]
                              for e in matching_temp}
        else:
            self._matching = {self._idx[e]: matching_temp[e] for e in matching_temp}

        # Build a DiGraph for doing basis exchange
        self._D = <GenericGraph_pyx?> DiGraph()
        # Make sure we get isolated vertices, corresponding to loops
        self._D.add_vertices(self._idx.values())
        # Also get isolated vertices corresponding to empty sets
        self._D.add_vertices(self._set_labels)

        # For sets in the matching, orient them as starting from the collections
        matching_reversed = ((v, k) for k, v in self._matching.items())
        self._D.add_edges(matching_reversed)

        other_edges = ((self._idx[e], set_labels[i])
                       for i, s in enumerate(sets)
                       for e in s
                       if e not in matching_temp or matching_temp[e] != set_labels[i])
        self._D.add_edges(other_edges)

    cdef bint _is_exchange_pair(self, long x, long y) except -1:
        r"""
        Check for `M`-alternating path from `x` to `y`.
        """
        return x in self._D.depth_first_search(y)

    cdef int _exchange(self, long x, long y) except -1:
        r"""
        Replace ``self.basis() with ``self.basis() - x + y``.

        Internal method, does no checks.
        """
        # update the internal matching
        sh = self._D.shortest_path(y, x)
        del self._matching[x]
        for i in range(0, len(sh)-1, 2):
            self._matching[sh[i]] = sh[i+1]

        # update the graph to reflect this new matching
        # i.e., reverse the orientation of edges along the shortest path sh
        self._D.delete_edges(zip(sh, sh[1:]))
        sh.reverse()
        self._D.add_edges(zip(sh, sh[1:]))

        BasisExchangeMatroid._exchange(self, x, y)

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: from sage.matroids.transversal_matroid import TransversalMatroid
            sage: sets = [[0, 1, 2, 3]] * 3
            sage: M = TransversalMatroid(sets); M
            Transversal matroid of rank 3 on 4 elements, with 3 sets
        """
        return "Transversal matroid of rank {} on {} elements, with {} sets".format(
            self.rank(), self.size(), len(self._sets_input))

    cpdef list sets(self):
        r"""
        Return the sets of ``self``.

        A transversal matroid can be viewed as a groundset with a collection
        from its powerset. This is represented as a bipartite graph, where
        an edge represents containment.

        OUTPUT: list of lists that correspond to the sets of the transversal matroid

        EXAMPLES::

            sage: from sage.matroids.transversal_matroid import TransversalMatroid
            sage: sets = [[0, 1, 2, 3], [1, 2], [3, 4]]
            sage: set_labels = [5, 6, 7]
            sage: M = TransversalMatroid(sets, set_labels=set_labels)
            sage: sorted(M.sets()) == sorted(sets)
            True
        """
        return copy(self._sets_input)

    def __richcmp__(left, right, op):
        r"""
        Compare two matroids.

        We take a very restricted view on equality: the objects need to be of
        the exact same type (so no subclassing) and the internal data need to
        be the same. For transversal matroids, in particular, the presentation
        as a bipartite graph must be the same.

        .. WARNING::

            This method is linked to ``__hash__``. If you override one, you
            MUST override the other!

        EXAMPLES::

            sage: from sage.matroids.transversal_matroid import TransversalMatroid
            sage: sets = [['a', 'b', 'c'], ['c', 'd'], ['a', 'd'], ['e']]
            sage: M = TransversalMatroid(sets)
            sage: M1 = TransversalMatroid(sets)
            sage: sets2 = [['a', 'b', 'c'], ['c', 'd'], ['a', 'd', 'e'], ['e']]
            sage: M2 = TransversalMatroid(sets2)
            sage: M1 == M2
            False
            sage: M1.equals(M2)
            True
            sage: from sage.matroids.transversal_matroid import TransversalMatroid
            sage: sets = [[0, 1, 2, 3], [1, 2], [1, 3, 4]]
            sage: M = TransversalMatroid(sets, groundset=range(5), set_labels=[5, 6, 7])
            sage: M2 = TransversalMatroid(sets)
            sage: M == M2
            True

        TESTS::

            sage: M = TransversalMatroid([range(5)] * 4, set_labels='abcd',
            ....:                        matching={0: 'a', 1: 'c'})
            sage: N = TransversalMatroid([range(5)] * 4, set_labels='abcd')
            sage: M == N
            False
        """
        if op not in [Py_EQ, Py_NE]:
            return NotImplemented
        if not isinstance(left, TransversalMatroid) or not isinstance(right, TransversalMatroid):
            return NotImplemented
        if left.__class__ != right.__class__:   # since we have some subclasses, an extra test
            return NotImplemented
        if op == Py_EQ:
            res = True
        if op == Py_NE:
            res = False
        # res gets inverted if matroids are deemed different.

        cdef TransversalMatroid rhs, lhs
        rhs = <TransversalMatroid> right
        lhs = <TransversalMatroid> left

        # If either matroid is not valid, deem them different
        if not (lhs.is_valid() and rhs.is_valid()):
            return not res
        if (lhs._groundset == rhs._groundset and
                Counter([frozenset(s) for s in lhs._sets_input]) == Counter([frozenset(s) for s in rhs._sets_input])):
            return res
        else:
            return not res

    def __hash__(self):
        r"""
        Return the hash of ``self``.

        The hash is based on the groundset and an internal Counter of the
        collection of sets.

        This function is called when matroids are added to a set. It is very
        desirable to override it so it can distinguish matroids on the same
        groundset, which is a very typical use case!

        .. WARNING::

            This method is linked to ``__richcmp__`` (in Cython) and
            ``__eq__`` (in Python). If you override one, you must
            override the other!

        EXAMPLES::

            sage: from sage.matroids.transversal_matroid import TransversalMatroid
            sage: sets1 = [[0, 1, 2, 3], [1, 2], [3, 4]]
            sage: M1 = TransversalMatroid(sets1)
            sage: M2 = TransversalMatroid(sets1, set_labels=[5, 6, 7])
            sage: sets3 = [['a', 'b', 'c'], ['c', 'd'], ['a', 'd', 'e']]
            sage: M3 = TransversalMatroid(sets3)
            sage: hash(M1) == hash(M2)
            True
            sage: hash(M1) == hash(M3)
            False
        """
        return hash((self._groundset, frozenset(self._sets.items())))

    cdef dict _translate_matching(self):
        """
        Return a Python dictionary that can be used as input for ``__init__()``.
        """
        cdef dict matching = {}
        for x in self._matching:
            matching[self._E[x]] = self._set_labels_input[self._set_labels.index(self._matching[x])]
        return matching

    def __reduce__(self):
        """
        Save the matroid for later reloading.

        OUTPUT:

        A tuple ``(unpickle, (version, data))``, where ``unpickle`` is the
        name of a function that, when called with ``(version, data)``,
        produces a matroid isomorphic to ``self``. ``version`` is an integer
        (currently 0) and ``data`` is a tuple ``(sets, E, name)`` where
        ``E`` is the groundset of the matroid, ``sets`` is the subsets of the
        transversal, and ``name`` is a custom name.

        EXAMPLES::

            sage: from sage.matroids.transversal_matroid import *
            sage: sets = [range(6)] * 3
            sage: M = TransversalMatroid(sets)
            sage: M == loads(dumps(M))
            True
            sage: M.rename('U36')
            sage: loads(dumps(M))
            U36
        """
        from sage.matroids.unpickling import unpickle_transversal_matroid
        data = (self._sets_input, self._E, self._set_labels_input, self._translate_matching(),
                self.get_custom_name())
        version = 0
        return unpickle_transversal_matroid, (version, data)

    def graph(self):
        """
        Return a bipartite graph representing the transversal matroid.

        A transversal matroid can be represented as a set system, or as a
        bipartite graph with one color class corresponding to the groundset
        and the other to the sets of the set system. This method returns
        that bipartite graph.

        OUTPUT: :class:`Graph`

        EXAMPLES::

            sage: from sage.matroids.transversal_matroid import TransversalMatroid
            sage: edgedict = {5: [0, 1, 2, 3], 6: [1, 2], 7: [1, 3, 4]}
            sage: B = BipartiteGraph(edgedict)
            sage: M = TransversalMatroid(edgedict.values(), set_labels=edgedict.keys())
            sage: M.graph() == B
            True
            sage: M2 = TransversalMatroid(edgedict.values())
            sage: B2 = M2.graph()
            sage: B2 == B
            False
            sage: B2.is_isomorphic(B)
            True
        """
        # relabel the vertices, then return as a BipartiteGraph
        vertex_map = dict(enumerate(self._E))
        for i, l in enumerate(self._set_labels):
            vertex_map[l] = self._set_labels_input[i]
        partition = [list(self._E), self._set_labels_input]

        return BipartiteGraph(self._D.relabel(vertex_map, inplace=False),
                              partition=partition)

    cpdef _minor(self, contractions, deletions):
        """
        Return a minor.

        Deletions will yield a new transversal matroid. Contractions will
        have to be a :class:`~sage.matroids.minor_matroid.MinorMatroid`
        until Gammoids are implemented.

        INPUT:

        - ``contractions`` -- frozenset; an independent subset of the groundset
        - ``deletions`` -- frozenset; a coindependent subset of the groundset

        OUTPUT:

        If ``contractions`` is the empty set, or if ``contractions``
        consists of only coloops,  an instance of
        :class:`~sage.matroids.transversal_matroid.TransversalMatroid`.
        Otherwise, an instance of
        :class:`~sage.matroids.minor_matroid.MinorMatroid`.

        EXAMPLES::

            sage: from sage.matroids.transversal_matroid import TransversalMatroid
            sage: sets = [[0, 1, 2, 3], [1, 2], [1, 3, 4]]
            sage: M1 = TransversalMatroid(sets)
            sage: N1 = M1.delete([2,3])
            sage: sets2 = [[0, 1], [1], [4]]
            sage: M2 = TransversalMatroid(sets2)
            sage: N1.is_isomorphic(M2)
            True
            sage: M1._minor(deletions=set([3]), contractions=set([4]))
            M / {4}, where M is Transversal matroid of rank 3 on 4 elements, with 3 sets

        ::

            sage: from sage.matroids.transversal_matroid import TransversalMatroid
            sage: sets = [['a', 'c'], ['e']]
            sage: gs = ['a', 'c', 'd', 'e']
            sage: M = TransversalMatroid(sets, groundset=gs)
            sage: N = M.delete(['d','e']); N
            Transversal matroid of rank 1 on 2 elements, with 1 sets

        TESTS::

            sage: from sage.matroids.transversal_matroid import TransversalMatroid
            sage: sets = [['a', 'c'], ['e']]
            sage: gs = ['a', 'c', 'd', 'e']
            sage: M = TransversalMatroid(sets, groundset=gs)
            sage: M.contract(['e'])
            Transversal matroid of rank 1 on 3 elements, with 1 sets
            sage: M.contract(['d', 'e'])
            Transversal matroid of rank 1 on 2 elements, with 1 sets
            sage: M.contract(['a', 'e'])
            M / {'a', 'e'}, where M is Transversal matroid of rank 2 on 4 elements, with 2 sets
        """
        cdef TransversalMatroid N
        cdef list new_sets, new_set_labels, s, new_s
        cdef size_t i
        cdef frozenset contract = frozenset(contractions)
        cdef frozenset delete = frozenset(deletions)

        # if contractions are just coloops, we can just delete them
        if self._corank(contract) == 0:
            delete = delete.union(contract)
            contract = frozenset()

        if delete:
            new_sets = []
            new_set_labels = []
            for i in range(len(self._sets_input)):
                s = <list> self._sets_input[i]
                new_s = [e for e in s if e not in delete]
                if new_s:
                    # skip over empty buckets, and do bookkeeping with the labels
                    new_sets.append(new_s)
                    new_set_labels.append(self._set_labels_input[i])
            groundset = self._groundset.difference(delete)
            N = TransversalMatroid(new_sets, groundset, new_set_labels)
            # Check if what remains is just coloops
            return N._minor(contractions=contract, deletions=frozenset())
        else:
            N = self

        if contractions:
            # Until gammoids are implemented
            return MinorMatroid(N, contractions=contract, deletions=frozenset())
        else:
            return N

    def set_labels(self):
        """
        Return the labels used for the transversal sets.

        This method will return a list of the labels used of the non-ground
        set vertices of the bipartite graph used to represent the transversal
        matroid. This method does not set anything.

        OUTPUT: list

        EXAMPLES::

            sage: from sage.matroids.transversal_matroid import TransversalMatroid
            sage: M = TransversalMatroid([[0, 1], [1, 2, 3], [3, 4, 7]])
            sage: M.set_labels()
            ['s0', 's1', 's2']
            sage: M.graph().vertices()
            ['s0', 's1', 's2', 0, 1, 2, 3, 4, 7]
        """
        return copy(self._set_labels_input)

    cpdef reduce_presentation(self):
        """
        Return an equal transversal matroid where the number of sets equals the rank.

        Every transversal matroid `M` has a presentation with `r(M)` sets, and if `M`
        has no coloops, then every presentation has `r(M)` nonempty sets. This method
        discards extra sets if `M` has coloops.

        OUTPUT: :class:`TransversalMatroid` with a reduced presentation

        EXAMPLES::

            sage: from sage.matroids.transversal_matroid import TransversalMatroid
            sage: sets = [[0, 1], [2], [2]]
            sage: M = TransversalMatroid(sets); M
            Transversal matroid of rank 2 on 3 elements, with 3 sets
            sage: N = M.reduce_presentation(); N
            Transversal matroid of rank 2 on 3 elements, with 2 sets
            sage: N.equals(M)
            True
            sage: N == M
            False
            sage: sets = [[0, 1], [], [], [2]]
            sage: M1 = TransversalMatroid(sets); M1
            Transversal matroid of rank 2 on 3 elements, with 4 sets
            sage: M1.reduce_presentation()
            Transversal matroid of rank 2 on 3 elements, with 2 sets

        ::

            sage: from sage.matroids.transversal_matroid import TransversalMatroid
            sage: sets = [[0, 1, 2, 3]] * 3
            sage: M = TransversalMatroid(sets)
            sage: N = M.reduce_presentation()
            sage: M == N
            True

        TESTS::

            sage: from sage.matroids.transversal_matroid import TransversalMatroid
            sage: sets = [[4], [1, 3], [4], [0, 1], [2, 3], [1]]
            sage: M = TransversalMatroid(sets)
            sage: M1 = M.reduce_presentation(); M1
            Transversal matroid of rank 5 on 5 elements, with 5 sets
            sage: len(M1.graph().edges())
            5
        """
        cdef TransversalMatroid N
        cdef size_t i
        if len(self._sets_input) == self.full_rank():
            return self

        element_int_map = {e: i for i, e in enumerate(self._groundset)}
        coloops = self.coloops()
        cdef list coloops_to_delete = [e for e in coloops if self._D.degree(element_int_map[e]) > 1]
        N = <TransversalMatroid?> self._minor(contractions=set(), deletions=set(coloops_to_delete))
        cdef list sets = list(N._sets_input)  # make a (shallow) copy as we will remove objects
        # reuse the old set labels
        # this does not respect containment
        labels = N.set_labels()
        # remove empty sets HERE
        if [] in sets:
            i = len(sets)
            while i > 0:
                i -= 1
                if not sets[i]:
                    del sets[i]
                    del labels[i]
        cdef set free_labels = set(self._set_labels_input).difference(labels)
        for c in coloops_to_delete:
            l = free_labels.pop()
            sets.append([c])
            labels.append(l)
        return TransversalMatroid(sets, groundset=self.groundset(), set_labels=labels)

    cpdef transversal_extension(self, element=None, newset=False, sets=None):
        r"""
        Return a :class:`TransversalMatroid` extended by an element.

        This will modify the presentation of the transversal matroid by
        adding a new element, and placing this element in the specified
        sets. It is also possible to use this method to create a new set
        that will have the new element as its only member, making it a coloop.

        INPUT:

        - ``element`` -- (optional) the name for the new element
        - ``newset`` -- (optional) if specified, the element will be
          given its own set
        - ``sets`` -- iterable of labels (default: ``None``) representing the
          sets in the current presentation that the new element will belong to

        OUTPUT:

        A :class:`~sage.matroids.transversal_matroids.TransversalMatroid`
        with a groundset element added to specified sets. Note that the
        ``newset`` option will make the new element a coloop. If
        ``newset == True``, a name will be generated; otherwise the
        value of ``newset`` will be used.

        EXAMPLES::

            sage: from sage.matroids.transversal_matroid import TransversalMatroid
            sage: M = TransversalMatroid([['a', 'c']], groundset=['a', 'c'], set_labels=['b'])
            sage: M1 = M.transversal_extension(element='d', newset='e')
            sage: M2 = M.transversal_extension(element='d', newset=True)
            sage: M1.coloops()
            frozenset({'d'})
            sage: True in M2.graph().vertices()
            False
            sage: M1.is_isomorphic(M2)
            True
            sage: M3 = M.transversal_extension('d', sets=['b'])
            sage: M3.is_isomorphic(matroids.Uniform(1, 3))
            True
            sage: M4 = M.transversal_extension('d', sets=['a'])
            Traceback (most recent call last):
            ...
            ValueError: sets do not match presentation
            sage: M4 = M.transversal_extension('a', sets=['b'])
            Traceback (most recent call last):
            ...
            ValueError: cannot extend by element already in groundset
            sage: M2.transversal_extension(newset='b')
            Traceback (most recent call last):
            ...
            ValueError: newset is already a vertex in the presentation
            sage: M5 = M1.transversal_extension()
            sage: len(M5.loops())
            1
            sage: M2.transversal_extension(element='b')
            Transversal matroid of rank 2 on 4 elements, with 2 sets

        ::

            sage: from sage.matroids.transversal_matroid import TransversalMatroid
            sage: sets = [[0, 1, 2, 3], [1, 2], [1, 3, 4]]
            sage: M = TransversalMatroid(sets, groundset=range(5), set_labels=[5, 6, 7])
            sage: N = M.delete(2)
            sage: M1 = N.transversal_extension(element=2, sets=[5, 6])
            sage: M1 == M
            True

        ::

            sage: from sage.matroids.transversal_matroid import TransversalMatroid
            sage: sets = [[0, 1, 2, 3]] * 3
            sage: M = TransversalMatroid(sets, set_labels=[4, 5, 6])
            sage: N = M.transversal_extension(element='a', newset=True, sets=[4])
            sage: N.graph().degree('a')
            2

        TESTS::

            sage: N = TransversalMatroid(['abc', 'abd', 'cde']); N
            Transversal matroid of rank 3 on 5 elements, with 3 sets
            sage: Ne = N.transversal_extension(element='f', sets=['s2'])
        """
        if sets is None:
            sets = []
        cdef set parsed_sets = set(sets)
        if element is None:
            element = newlabel(self._groundset)
        elif element in self._groundset:
            raise ValueError("cannot extend by element already in groundset")
        cdef list labels = self._set_labels_input
        if not parsed_sets.issubset(labels):
            raise ValueError("sets do not match presentation")

        # check for conflicts with new labels
        labels_map = {l: l for l in labels}
        if element in labels:
            new_label = newlabel(self._groundset.union(labels).union([newset]))
            labels_map[element] = new_label

        # newset should not be a groundset element or existing set
        if newset in self._E or newset in self._set_labels_input:
            # keywords `True` and `False` give us problems here
            if not isinstance(newset, bool):
                raise ValueError("newset is already a vertex in the presentation")

        cdef list s, new_sets = []
        cdef size_t i
        for i in range(len(self._sets_input)):
            s = <list> self._sets_input[i]
            if labels[i] in parsed_sets:
                new_sets.append(s + [element])
            else:
                new_sets.append(s)

        groundset = self._groundset.union([element])
        if newset:
            if newset is True:
                newset = newlabel(groundset.union(labels))
            new_sets.append([element])
            labels = list(labels)  # Make a shallow copy since we mutate it
            labels.append(newset)

        return TransversalMatroid(new_sets, groundset, labels)

    def transversal_extensions(self, element=None, sets=None):
        r"""
        Return an iterator of extensions based on the transversal presentation.

        This method will take an extension by adding an element to every possible
        sub-collection of the collection of desired sets. No checking is done
        for equal matroids. It is advised to make sure the presentation has as
        few sets as possible by using
        :meth:`reduce_presentation() <sage.matroids.transversal_matroid.TransversalMatroid.reduce_presentation>`

        INPUT:

        - ``element`` -- (optional) the name of the new element
        - ``sets`` -- (optional) list containing names of sets in the matroid's
          presentation

        OUTPUT: iterator over instances of :class:`TransversalMatroid`

        If ``sets`` is not specified, every set will be used.

        EXAMPLES::

            sage: from sage.matroids.transversal_matroid import TransversalMatroid
            sage: sets = [[3, 4, 5, 6]] * 3
            sage: M = TransversalMatroid(sets, set_labels=[0, 1, 2])
            sage: len([N for N in M.transversal_extensions()])
            8
            sage: len([N for N in M.transversal_extensions(sets=[0, 1])])
            4
            sage: set(sorted([N.groundset() for N in M.transversal_extensions(element=7)]))
            {frozenset({3, 4, 5, 6, 7})}
        """
        if element is None:
            element = newlabel(self.groundset())
        elif element in self._groundset:
            raise ValueError("cannot extend by element already in groundset")

        labels = self._set_labels_input
        if sets is None:
            sets = labels
        elif not set(sets).issubset(labels):
            raise ValueError("sets do not match presentation")

        # Adapted from the Python documentation
        from itertools import chain, combinations
        powerset = chain.from_iterable(combinations(sets, r) for r in range(len(sets)+1))

        for collection in powerset:
            yield self.transversal_extension(element=element, sets=collection)

    cpdef is_valid(self, certificate=False):
        """
        Test whether the matching in memory is a valid maximal matching.

        The data for a transversal matroid is a set system, which is always valid,
        but it is possible for a user to provide invalid input with the ``matching``
        parameter. This checks that the matching provided is indeed a matching, fits in
        the set system, and is maximal.

        EXAMPLES::

            sage: from sage.matroids.transversal_matroid import *
            sage: sets = [[0, 1, 2, 3], [1, 2], [1, 3, 4]]
            sage: set_labels = [5, 6, 7]
            sage: M = TransversalMatroid(sets, set_labels=set_labels)
            sage: M.is_valid()
            True
            sage: m = {0: 5, 1: 5, 3: 7}  # not a matching
            sage: TransversalMatroid(sets, set_labels=set_labels, matching=m).is_valid()
            False
            sage: m = {2: 6, 3: 7}  # not maximal
            sage: TransversalMatroid(sets, set_labels=set_labels, matching=m).is_valid()
            False
            sage: m = {0: 6, 1: 5, 3: 7}  # not in the set system
            sage: TransversalMatroid(sets, set_labels=set_labels, matching=m).is_valid()
            False
        """
        # Check that self._matching is a matching, that is, every entry appears at most once.
        # Because of how python dictionaries work, each element appears once, but a set
        # can appear more than once.
        if len(self._matching) != len(set(self._matching.values())):
            return False

        # Check that every element in the matching is actually in the set that the matching
        # specifies.
        # Note that self._matching uses the internal set and element labels
        for e in self._matching:
            if e not in self._sets_input[self._set_labels.index(self._matching[e])]:
                return False

        # Check that the matching is maximal.
        B = self.graph()
        if len(B.matching()) != len(self._matching):
            return False

        return True



"""
Weight lattices and weight spaces
"""
# ****************************************************************************
#       Copyright (C) 2008-2009 Nicolas M. Thiery <nthiery at users.sf.net>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.misc.cachefunc import cached_method
from sage.sets.family import Family
from sage.combinat.free_module import CombinatorialFreeModule
from .weight_lattice_realizations import WeightLatticeRealizations
import functools


class WeightSpace(CombinatorialFreeModule):
    r"""
    INPUT:

    - ``root_system`` -- a root system
    - ``base_ring`` -- a ring `R`
    - ``extended`` -- boolean (default: ``False``)

    The weight space (or lattice if ``base_ring`` is `\ZZ`) of a root
    system is the formal free module `\bigoplus_i R \Lambda_i`
    generated by the fundamental weights `(\Lambda_i)_{i\in I}` of the
    root system.

    This class is also used for coweight spaces (or lattices).

    .. SEEALSO::

        - :meth:`RootSystem`
        - :meth:`RootSystem.weight_lattice` and :meth:`RootSystem.weight_space`
        - :meth:`~sage.combinat.root_system.weight_lattice_realizations.WeightLatticeRealizations`

    EXAMPLES::

        sage: Q = RootSystem(['A', 3]).weight_lattice(); Q
        Weight lattice of the Root system of type ['A', 3]
        sage: Q.simple_roots()                                                          # needs sage.graphs
        Finite family {1: 2*Lambda[1] -   Lambda[2],
                       2:  -Lambda[1] + 2*Lambda[2] -   Lambda[3],
                       3:                -Lambda[2] + 2*Lambda[3]}

        sage: Q = RootSystem(['A', 3, 1]).weight_lattice(); Q
        Weight lattice of the Root system of type ['A', 3, 1]
        sage: Q.simple_roots()                                                          # needs sage.graphs
        Finite family {0: 2*Lambda[0] -   Lambda[1]               -   Lambda[3],
                       1:  -Lambda[0] + 2*Lambda[1] -   Lambda[2],
                       2:                -Lambda[1] + 2*Lambda[2] -   Lambda[3],
                       3:  -Lambda[0]               -   Lambda[2] + 2*Lambda[3]}

    For infinite types, the Cartan matrix is singular, and therefore
    the embedding of the root lattice is not faithful::

        sage: sum(Q.simple_roots())                                                     # needs sage.graphs
        0

    In particular, the null root is zero::

        sage: Q.null_root()                                                             # needs sage.graphs
        0

    This can be compensated by extending the basis of the weight space
    and slightly deforming the simple roots to make them linearly
    independent, without affecting the scalar product with the
    coroots. This feature is currently only implemented for affine
    types. In that case, if ``extended`` is set, then the basis of the
    weight space is extended by an element `\delta`::

        sage: Q = RootSystem(['A', 3, 1]).weight_lattice(extended=True); Q
        Extended weight lattice of the Root system of type ['A', 3, 1]
        sage: Q.basis().keys()
        {0, 1, 2, 3, 'delta'}

    And the simple root `\alpha_0` associated to the special node is
    deformed as follows::

        sage: Q.simple_roots()                                                          # needs sage.graphs
        Finite family {0: 2*Lambda[0] -   Lambda[1]               -   Lambda[3] + delta,
                       1:  -Lambda[0] + 2*Lambda[1] -   Lambda[2],
                       2:                -Lambda[1] + 2*Lambda[2] -   Lambda[3],
                       3:  -Lambda[0]               -   Lambda[2] + 2*Lambda[3]}

    Now, the null root is nonzero::

        sage: Q.null_root()                                                             # needs sage.graphs
        delta

    .. WARNING::

        By a slight notational abuse, the extra basis element used to
        extend the fundamental weights is called ``\delta`` in the
        current implementation. However, in the literature,
        ``\delta`` usually denotes instead the null root. Most of the
        time, those two objects coincide, but not for type `BC` (aka.
        `A_{2n}^{(2)}`). Therefore we currently have::

            sage: Q = RootSystem(["A",4,2]).weight_lattice(extended=True)
            sage: Q.simple_root(0)                                                      # needs sage.graphs
            2*Lambda[0] - Lambda[1] + delta
            sage: Q.null_root()                                                         # needs sage.graphs
            2*delta

        whereas, with the standard notations from the literature, one
        would expect to get respectively `2\Lambda_0 -\Lambda_1 +1/2
        \delta` and `\delta`.

        Other than this notational glitch, the implementation remains
        correct for type `BC`.

        The notations may get improved in a subsequent version, which
        might require changing the index of the extra basis
        element. To guarantee backward compatibility in code not
        included in Sage, it is recommended to use the following idiom
        to get that index::

            sage: F = Q.basis_extension(); F
            Finite family {'delta': delta}
            sage: index = F.keys()[0]; index
            'delta'

        Then, for example, the coefficient of an element of the
        extended weight lattice on that basis element can be recovered
        with::

            sage: Q.null_root()[index]                                                  # needs sage.graphs
            2

    TESTS::

        sage: for ct in (CartanType.samples(crystallographic=True)                      # needs sage.graphs
        ....:             + [CartanType(["A",2], ["C",5,1])]):
        ....:     TestSuite(ct.root_system().weight_lattice()).run()
        ....:     TestSuite(ct.root_system().weight_space()).run()
        sage: for ct in CartanType.samples(affine=True):                                # needs sage.graphs
        ....:     if ct.is_implemented():
        ....:         P = ct.root_system().weight_space(extended=True)
        ....:         TestSuite(P).run()
    """

    @staticmethod
    def __classcall_private__(cls, root_system, base_ring, extended=False):
        """
        Guarantee Unique representation.

        .. SEEALSO:: :class:`UniqueRepresentation`

        TESTS::

            sage: R = RootSystem(['A',4])
            sage: from sage.combinat.root_system.weight_space import WeightSpace
            sage: WeightSpace(R, QQ) is WeightSpace(R, QQ, False)
            True
        """
        return super().__classcall__(cls, root_system, base_ring, extended)

    def __init__(self, root_system, base_ring, extended):
        """
        TESTS::

            sage: R = RootSystem(['A',4])
            sage: from sage.combinat.root_system.weight_space import WeightSpace
            sage: Q = WeightSpace(R, QQ); Q
            Weight space over the Rational Field of the Root system of type ['A', 4]
            sage: TestSuite(Q).run()                                                    # needs sage.graphs

            sage: WeightSpace(R, QQ, extended=True)
            Traceback (most recent call last):
            ...
            ValueError: extended weight lattices are only implemented for affine root systems
        """
        basis_keys = root_system.index_set()
        self._extended = extended
        if extended:
            if not root_system.cartan_type().is_affine():
                raise ValueError("extended weight lattices are only"
                                 " implemented for affine root systems")
            basis_keys = tuple(basis_keys) + ("delta",)

            def sortkey(x):
                return (1 if isinstance(x, str) else 0, x)
        else:
            def sortkey(x):
                return x

        self.root_system = root_system
        CombinatorialFreeModule.__init__(self, base_ring,
                                         basis_keys,
                                         prefix="Lambdacheck" if root_system.dual_side else "Lambda",
                                         latex_prefix="\\Lambda^\\vee" if root_system.dual_side else "\\Lambda",
                                         sorting_key=sortkey,
                                         category=WeightLatticeRealizations(base_ring))

        if root_system.cartan_type().is_affine() and not extended:
            # For an affine type, register the quotient map from the
            # extended weight lattice/space to the weight lattice/space
            domain = root_system.weight_space(base_ring, extended=True)
            domain.module_morphism(self.fundamental_weight,
                                   codomain=self
                                   ).register_as_coercion()

    def is_extended(self):
        """
        Return whether this is an extended weight lattice.

        .. SEEALSO:: :meth:`~sage.combinat.root_system.weight_lattice_realization.ParentMethods.is_extended`

        EXAMPLES::

            sage: RootSystem(["A",3,1]).weight_lattice().is_extended()
            False
            sage: RootSystem(["A",3,1]).weight_lattice(extended=True).is_extended()
            True
        """
        return self._extended

    def _repr_(self):
        """
        TESTS::

            sage: RootSystem(['A',4]).weight_lattice()    # indirect doctest
            Weight lattice of the Root system of type ['A', 4]
            sage: RootSystem(['B',4]).weight_space()
            Weight space over the Rational Field of the Root system of type ['B', 4]
            sage: RootSystem(['A',4]).coweight_lattice()
            Coweight lattice of the Root system of type ['A', 4]
            sage: RootSystem(['B',4]).coweight_space()
            Coweight space over the Rational Field of the Root system of type ['B', 4]
        """
        return self._name_string()

    def _name_string(self, capitalize=True, base_ring=True, type=True):
        """
        EXAMPLES::

            sage: RootSystem(['A',4]).weight_lattice()._name_string()
            "Weight lattice of the Root system of type ['A', 4]"
        """
        return self._name_string_helper("weight",
                                        capitalize=capitalize, base_ring=base_ring, type=type,
                                        prefix="extended " if self.is_extended() else "")

    @cached_method
    def fundamental_weight(self, i):
        r"""
        Return the `i`-th fundamental weight.

        INPUT:

        - ``i`` -- an element of the index set or ``'delta'``

        By a slight notational abuse, for an affine type this method
        also accepts ``'delta'`` as input, and returns the image of
        `\delta` of the extended weight lattice in this realization.

        .. SEEALSO:: :meth:`~sage.combinat.root_system.weight_lattice_realization.ParentMethods.fundamental_weight`

        EXAMPLES::

            sage: Q = RootSystem(["A",3]).weight_lattice()
            sage: Q.fundamental_weight(1)
            Lambda[1]

            sage: Q = RootSystem(["A",3,1]).weight_lattice(extended=True)
            sage: Q.fundamental_weight(1)
            Lambda[1]
            sage: Q.fundamental_weight("delta")
            delta
        """
        if i == "delta":
            if not self.cartan_type().is_affine():
                raise ValueError("delta is only defined for affine weight spaces")
            if self.is_extended():
                return self.monomial(i)
            return self.zero()
        if i not in self.index_set():
            raise ValueError("{} is not in the index set".format(i))
        return self.monomial(i)

    @cached_method
    def basis_extension(self):
        r"""
        Return the basis elements used to extend the fundamental weights.

        EXAMPLES::

            sage: Q = RootSystem(["A",3,1]).weight_lattice()
            sage: Q.basis_extension()
            Family ()

            sage: Q = RootSystem(["A",3,1]).weight_lattice(extended=True)
            sage: Q.basis_extension()
            Finite family {'delta': delta}

        This method is irrelevant for finite types::

            sage: Q = RootSystem(["A",3]).weight_lattice()
            sage: Q.basis_extension()
            Family ()
        """
        if self.is_extended():
            return Family(["delta"], self.monomial)
        return Family([])

    @cached_method
    def simple_root(self, j):
        r"""
        Return the `j`-th simple root.

        EXAMPLES::

            sage: L = RootSystem(["C",4]).weight_lattice()
            sage: L.simple_root(3)                                                      # needs sage.graphs
            -Lambda[2] + 2*Lambda[3] - Lambda[4]

        Its coefficients are given by the corresponding column of the
        Cartan matrix::

            sage: L.cartan_type().cartan_matrix()[:,2]                                  # needs sage.graphs
            [ 0]
            [-1]
            [ 2]
            [-1]

        Here are all simple roots::

            sage: L.simple_roots()                                                      # needs sage.graphs
            Finite family {1:  2*Lambda[1]   - Lambda[2],
                           2:   -Lambda[1] + 2*Lambda[2]   - Lambda[3],
                           3:   -Lambda[2] + 2*Lambda[3]   - Lambda[4],
                           4:               -2*Lambda[3] + 2*Lambda[4]}

        For the extended weight lattice of an affine type, the simple
        root associated to the special node is deformed by adding
        `\delta`, where `\delta` is the null root::

            sage: L = RootSystem(["C",4,1]).weight_lattice(extended=True)
            sage: L.simple_root(0)                                                      # needs sage.graphs
            2*Lambda[0] - 2*Lambda[1] + delta

        In fact `\delta` is really `1/a_0` times the null root (see
        the discussion in :class:`~sage.combinat.root_system.weight_space.WeightSpace`)
        but this only makes a difference in type `BC`::

            sage: L = RootSystem(CartanType(["BC",4,2])).weight_lattice(extended=True)
            sage: L.simple_root(0)                                                      # needs sage.graphs
            2*Lambda[0] - Lambda[1] + delta
            sage: L.null_root()                                                         # needs sage.graphs
            2*delta

        .. SEEALSO::

            - :meth:`~sage.combinat.root_system.type_affine.AmbientSpace.simple_root`
            - :meth:`CartanType.col_annihilator`
        """
        if j not in self.index_set():
            raise ValueError("{} is not in the index set".format(j))
        K = self.base_ring()
        result = self.sum_of_terms((i,K(c)) for i,c in self.root_system.dynkin_diagram().column(j))
        if self._extended and j == self.cartan_type().special_node():
            result = result + self.monomial("delta")
        return result

    def _repr_term(self, m):
        r"""
        Customized monomial printing for extended weight lattices.

        EXAMPLES::

            sage: L = RootSystem(["C",4,1]).weight_lattice(extended=True)
            sage: L.simple_root(0)             # indirect doctest                       # needs sage.graphs
            2*Lambda[0] - 2*Lambda[1] + delta

            sage: L = RootSystem(["C",4,1]).coweight_lattice(extended=True)
            sage: L.simple_root(0)             # indirect doctest                       # needs sage.graphs
            2*Lambdacheck[0] - Lambdacheck[1] + deltacheck
        """
        if m == "delta":
            return "deltacheck" if self.root_system.dual_side else "delta"
        return super()._repr_term(m)

    def _latex_term(self, m):
        r"""
        Customized monomial typesetting for extended weight lattices.

        EXAMPLES::

            sage: L = RootSystem(["C",4,1]).weight_lattice(extended=True)
            sage: latex(L.simple_root(0))             # indirect doctest                # needs sage.graphs
            2 \Lambda_{0} - 2 \Lambda_{1} + \delta

            sage: L = RootSystem(["C",4,1]).coweight_lattice(extended=True)
            sage: latex(L.simple_root(0))             # indirect doctest                # needs sage.graphs
            2 \Lambda^\vee_{0} - \Lambda^\vee_{1} + \delta^\vee
        """
        if m == "delta":
            return "\\delta^\\vee" if self.root_system.dual_side else "\\delta"
        return super()._latex_term(m)

    @cached_method
    def _to_classical_on_basis(self, i):
        r"""
        Implement the projection onto the corresponding classical space or lattice, on the basis.

        INPUT:

        - ``i`` -- a vertex of the Dynkin diagram or "delta"

        EXAMPLES::

            sage: L = RootSystem(["A",2,1]).weight_space()
            sage: L._to_classical_on_basis("delta")
            0
            sage: L._to_classical_on_basis(0)
            0
            sage: L._to_classical_on_basis(1)
            Lambda[1]
            sage: L._to_classical_on_basis(2)
            Lambda[2]
        """
        if i == "delta" or i == self.cartan_type().special_node():
            return self.classical().zero()
        return self.classical().monomial(i)

    @cached_method
    def to_ambient_space_morphism(self):
        r"""
        The morphism from ``self`` to its associated ambient space.

        EXAMPLES::

            sage: CartanType(['A',2]).root_system().weight_lattice().to_ambient_space_morphism()
            Generic morphism:
            From: Weight lattice of the Root system of type ['A', 2]
            To:   Ambient space of the Root system of type ['A', 2]

        .. warning::

            Implemented only for finite Cartan type.
        """
        if self.root_system.dual_side:
            raise TypeError("No implemented map from the coweight space to the ambient space")
        L = self.cartan_type().root_system().ambient_space()
        basis = L.fundamental_weights()

        def basis_value(basis, i):
            return basis[i]
        return self.module_morphism(on_basis=functools.partial(basis_value, basis), codomain=L)


class WeightSpaceElement(CombinatorialFreeModule.Element):

    def scalar(self, lambdacheck):
        """
        The canonical scalar product between the weight lattice and
        the coroot lattice.

        .. TODO::

            - merge with_apply_multi_module_morphism
            - allow for any root space / lattice
            - define properly the return type (depends on the base rings of the two spaces)
            - make this robust for extended weight lattices (`i` might be "delta")

        EXAMPLES::

            sage: L = RootSystem(["C",4,1]).weight_lattice()
            sage: Lambda     = L.fundamental_weights()
            sage: alphacheck = L.simple_coroots()
            sage: Lambda[1].scalar(alphacheck[1])
            1
            sage: Lambda[1].scalar(alphacheck[2])
            0

        The fundamental weights and the simple coroots are dual bases::

            sage: matrix([ [ Lambda[i].scalar(alphacheck[j])
            ....:            for i in L.index_set() ]
            ....:          for j in L.index_set() ])
            [1 0 0 0 0]
            [0 1 0 0 0]
            [0 0 1 0 0]
            [0 0 0 1 0]
            [0 0 0 0 1]

        Note that the scalar product is not yet implemented between
        the weight space and the coweight space; in any cases, that
        won't be the job of this method::

            sage: R = RootSystem(["A",3])
            sage: alpha = R.weight_space().roots()                                      # needs sage.graphs
            sage: alphacheck = R.coweight_space().roots()                               # needs sage.graphs
            sage: alpha[1].scalar(alphacheck[1])                                        # needs sage.graphs
            Traceback (most recent call last):
            ...
            ValueError: -Lambdacheck[1] + 2*Lambdacheck[2] - Lambdacheck[3]
            is not in the coroot space
        """
        # TODO: Find some better test
        if lambdacheck not in self.parent().coroot_lattice() and lambdacheck not in self.parent().coroot_space():
            raise ValueError("{} is not in the coroot space".format(lambdacheck))
        zero = self.parent().base_ring().zero()
        if len(self) < len(lambdacheck):
            return sum( (lambdacheck[i]*c for (i,c) in self), zero)
        return sum( (self[i]*c for (i,c) in lambdacheck), zero)

    def is_dominant(self):
        r"""
        Check whether an element in the weight space lies in the positive
        cone spanned by the basis elements (fundamental weights).

        EXAMPLES::

            sage: W = RootSystem(['A',3]).weight_space()
            sage: Lambda = W.basis()
            sage: w = Lambda[1] + Lambda[3]
            sage: w.is_dominant()
            True
            sage: w = Lambda[1] - Lambda[2]
            sage: w.is_dominant()
            False

        In the extended affine weight lattice, ``'delta'`` is orthogonal to
        the positive coroots, so adding or subtracting it should not
        affect dominance::

            sage: P = RootSystem(['A',2,1]).weight_lattice(extended=true)
            sage: Lambda = P.fundamental_weights()
            sage: delta = P.null_root()                                                 # needs sage.graphs
            sage: w = Lambda[1] - delta                                                 # needs sage.graphs
            sage: w.is_dominant()                                                       # needs sage.graphs
            True
        """
        index_set = set(self.parent().index_set())
        return all(c >= 0 for i, c in self._monomial_coefficients.items() if i in index_set)

    def is_dominant_weight(self):
        r"""
        Check whether an element in the weight space lies in the positive
        `\ZZ`-lattice cone spanned by the basis elements (fundamental weights).

        EXAMPLES::

            sage: W = RootSystem(['A',3]).weight_space()
            sage: Lambda = W.basis()
            sage: w = Lambda[1] + Lambda[3]
            sage: w.is_dominant_weight()
            True
            sage: w = Lambda[1] + 2/3*Lambda[3]
            sage: w.is_dominant_weight()
            False
            sage: w = Lambda[1] - Lambda[2]
            sage: w.is_dominant_weight()
            False

        In the extended affine weight lattice, ``'delta'`` is orthogonal to
        the positive coroots, so adding or subtracting it should not
        affect dominance::

            sage: P = RootSystem(['A',2,1]).weight_lattice(extended=true)
            sage: Lambda = P.fundamental_weights()
            sage: delta = P.null_root()                                                 # needs sage.graphs
            sage: w = Lambda[1] - delta                                                 # needs sage.graphs
            sage: w.is_dominant_weight()                                                # needs sage.graphs
            True
        """
        index_set = set(self.parent().index_set())
        from sage.rings.integer_ring import ZZ
        return all(c in ZZ and c >= 0 for i, c in self._monomial_coefficients.items() if i in index_set)

    def to_ambient(self):
        r"""
        Map ``self`` to the ambient space.

        EXAMPLES::

            sage: mu = CartanType(['B',2]).root_system().weight_lattice().an_element(); mu
            2*Lambda[1] + 2*Lambda[2]
            sage: mu.to_ambient()
            (3, 1)

        .. WARNING::

            Only implemented in finite Cartan type.
            Does not work for coweight lattices because there is no implemented map
            from the coweight lattice to the ambient space.
        """
        return self.parent().to_ambient_space_morphism()(self)

    def to_weight_space(self):
        r"""
        Map ``self`` to the weight space.

        Since `self.parent()` is the weight space, this map just returns ``self``.
        This overrides the generic method in `WeightSpaceRealizations`.

        EXAMPLES::

            sage: mu = CartanType(['A',2]).root_system().weight_lattice().an_element(); mu
            2*Lambda[1] + 2*Lambda[2]
            sage: mu.to_weight_space()
            2*Lambda[1] + 2*Lambda[2]
        """
        return self

    @cached_method
    def _to_root_vector(self):
        r"""
        Helper method to express ``self`` as a linear combination
        of simple roots.

        OUTPUT:

        A vector with entries in `\QQ` representing ``self`` as a linear
        combination of simple roots.

        EXAMPLES::

            sage: P = RootSystem(['A',3]).weight_lattice()
            sage: La = P.fundamental_weights()
            sage: [al._to_root_vector() for al in P.simple_roots()]
            [(1, 0, 0), (0, 1, 0), (0, 0, 1)]
            sage: (La[1] + La[2])._to_root_vector()
            (5/4, 3/2, 3/4)
        """
        return self.parent()._inverse_cartan_matrix * self.to_vector()


WeightSpace.Element = WeightSpaceElement
